from dymos.models.atmosphere.atmos_1976 import USatm1976Comp

from aviary.mission.gasp_based.flight_conditions import FlightConditions
from aviary.mission.gasp_based.ode.base_ode import BaseODE
from aviary.mission.gasp_based.ode.params import ParamPort
from aviary.mission.gasp_based.phases.landing_components import (
    GlideConditionComponent, LandingAltitudeComponent,
    LandingGroundRollComponent)
from aviary.subsystems.aerodynamics.gasp_based.gaspaero import LowSpeedAero
from aviary.subsystems.propulsion.propulsion_mission import PropulsionMission
from aviary.variable_info.enums import SpeedType
from aviary.variable_info.variables import Aircraft, Dynamic, Mission


class LandingSegment(BaseODE):
    def setup(self):
        # TODO: paramport
        self.add_subsystem("params", ParamPort(), promotes=["*"])

        self.add_subsystem(
            "approach_alt_comp",
            LandingAltitudeComponent(),
            promotes_inputs=[
                Mission.Landing.OBSTACLE_HEIGHT,
                Mission.Landing.AIRPORT_ALTITUDE,
            ],
            promotes_outputs=[Mission.Landing.INITIAL_ALTITUDE],
        )

        self.add_subsystem(
            "USatm_app",
            USatm1976Comp(num_nodes=1),
            promotes_inputs=[("h", Mission.Landing.INITIAL_ALTITUDE)],
            promotes_outputs=[
                ("rho", "rho_app"),
                ("sos", "sos_app"),
                ("temp", "T_app"),
                ("pres", "P_app"),
                ("viscosity", "viscosity_app"),
            ],
        )

        self.add_subsystem(
            "fc_app",
            FlightConditions(num_nodes=1, input_speed_type=SpeedType.MACH),
            promotes_inputs=[
                ("rho", "rho_app"),
                (Dynamic.Mission.SPEED_OF_SOUND, "sos_app"),
                (Dynamic.Mission.MACH, Mission.Landing.INITIAL_MACH),
            ],
            promotes_outputs=[(Dynamic.Mission.DYNAMIC_PRESSURE, "q_app")],
        )

        propulsion_mission = self.add_subsystem(
            name='propulsion',
            subsys=PropulsionMission(
                num_nodes=1,
                aviary_options=self.options['aviary_options']),
            promotes_inputs=["*", (Dynamic.Mission.ALTITUDE, Mission.Landing.INITIAL_ALTITUDE),
                             (Dynamic.Mission.MACH, Mission.Landing.INITIAL_MACH)],
            promotes_outputs=[(Dynamic.Mission.THRUST_TOTAL, "thrust_idle")])
        propulsion_mission.set_input_defaults(Dynamic.Mission.THROTTLE, 0.0)

        # alpha input not needed, only used for CL_max
        self.add_subsystem(
            "aero_app",
            LowSpeedAero(num_nodes=1),
            promotes_inputs=[
                "*",
                (Dynamic.Mission.ALTITUDE, Mission.Landing.INITIAL_ALTITUDE),
                ("rho", "rho_app"),
                (Dynamic.Mission.SPEED_OF_SOUND, "sos_app"),
                ("viscosity", "viscosity_app"),
                ("airport_alt", Mission.Landing.AIRPORT_ALTITUDE),
                (Dynamic.Mission.MACH, Mission.Landing.INITIAL_MACH),
                (Dynamic.Mission.DYNAMIC_PRESSURE, "q_app"),
                ("flap_defl", Aircraft.Wing.FLAP_DEFLECTION_LANDING),
                ("t_init_flaps", "t_init_flaps_app"),
                ("t_init_gear", "t_init_gear_app"),
                ("CL_max_flaps", Mission.Landing.LIFT_COEFFICIENT_MAX),
                (
                    "dCL_flaps_model",
                    Mission.Landing.LIFT_COEFFICIENT_FLAP_INCREMENT,
                ),
                (
                    "dCD_flaps_model",
                    Mission.Landing.DRAG_COEFFICIENT_FLAP_INCREMENT,
                ),
            ],
            promotes_outputs=["CL_max"],
        )

        self.add_subsystem(
            "glide",
            GlideConditionComponent(),
            promotes_inputs=[
                "rho_app",
                Mission.Landing.MAXIMUM_SINK_RATE,
                Dynamic.Mission.MASS,
                Aircraft.Wing.AREA,
                Mission.Landing.GLIDE_TO_STALL_RATIO,
                "CL_max",
                Mission.Landing.MAXIMUM_FLARE_LOAD_FACTOR,
                Mission.Landing.TOUCHDOWN_SINK_RATE,
                Mission.Landing.INITIAL_ALTITUDE,
                Mission.Landing.BRAKING_DELAY,
            ],
            promotes_outputs=[
                Mission.Landing.INITIAL_VELOCITY,
                Mission.Landing.STALL_VELOCITY,
                "TAS_touchdown",
                "density_ratio",
                "wing_loading_land",
                "theta",
                "glide_distance",
                "tr_distance",
                "delay_distance",
                "flare_alt",
            ],
        )

        self.add_subsystem(
            "USatm_td",
            USatm1976Comp(num_nodes=1),
            promotes_inputs=[("h", Mission.Landing.AIRPORT_ALTITUDE)],
            promotes_outputs=[
                ("rho", "rho_td"),
                ("sos", "sos_td"),
                ("viscosity", "viscosity_td"),
            ],
        )

        self.add_subsystem(
            "fc_td",
            FlightConditions(num_nodes=1),
            promotes_inputs=[
                ("rho", "rho_td"),
                (Dynamic.Mission.SPEED_OF_SOUND, "sos_td"),
                ("TAS", "TAS_touchdown"),
            ],
            promotes_outputs=[(Dynamic.Mission.DYNAMIC_PRESSURE, "q_td"),
                              (Dynamic.Mission.MACH, "mach_td")],
        )

        self.add_subsystem(
            "aero_td",
            LowSpeedAero(num_nodes=1, retract_flaps=True, retract_gear=False),
            promotes_inputs=[
                "*",
                (Dynamic.Mission.ALTITUDE, Mission.Landing.AIRPORT_ALTITUDE),
                ("rho", "rho_td"),
                (Dynamic.Mission.SPEED_OF_SOUND, "sos_td"),
                ("viscosity", "viscosity_td"),
                ("airport_alt", Mission.Landing.AIRPORT_ALTITUDE),
                (Dynamic.Mission.MACH, "mach_td"),
                (Dynamic.Mission.DYNAMIC_PRESSURE, "q_td"),
                ("alpha", Aircraft.Wing.INCIDENCE),
                ("flap_defl", Aircraft.Wing.FLAP_DEFLECTION_LANDING),
                ("CL_max_flaps", Mission.Landing.LIFT_COEFFICIENT_MAX),
                (
                    "dCL_flaps_model",
                    Mission.Landing.LIFT_COEFFICIENT_FLAP_INCREMENT,
                ),
                (
                    "dCD_flaps_model",
                    Mission.Landing.DRAG_COEFFICIENT_FLAP_INCREMENT,
                ),
                ("t_init_flaps", "t_init_flaps_td"),
                ("t_init_gear", "t_init_gear_td"),
            ],
            promotes_outputs=[("CD", "touchdown_CD"), ("CL", "touchdown_CL")],
        )
        # GASP seems to run groundroll with flaps up and gear down (IWLD=2)
        self.set_input_defaults("t_init_flaps_td", 1e10)  # never deploy
        self.set_input_defaults("t_init_gear_td", 1e10)  # ensure gear down

        self.add_subsystem(
            "landinggroundroll",
            LandingGroundRollComponent(),
            promotes_inputs=[
                "touchdown_CD",
                "touchdown_CL",
                "TAS_touchdown",
                "thrust_idle",
                "density_ratio",
                "wing_loading_land",
                "glide_distance",
                "tr_distance",
                "delay_distance",
                "CL_max",
                Dynamic.Mission.MASS,
                'mission:*'
            ],
            promotes_outputs=[
                "ground_roll_distance", "average_acceleration", 'mission:*'],
        )

        ParamPort.set_default_vals(self)
        self.set_input_defaults(Mission.Landing.INITIAL_MACH, val=0.1)
        # landing doesn't change flap or gear position
        self.set_input_defaults("t_init_flaps_app", val=1e10)
        self.set_input_defaults("t_init_gear_app", val=1e10)
        self.set_input_defaults(
            Mission.Landing.INITIAL_ALTITUDE, val=50, units="ft")
        self.set_input_defaults('aero_ramps.flap_factor:final_val', val=1.)
        self.set_input_defaults('aero_ramps.gear_factor:final_val', val=1.)
        self.set_input_defaults('aero_ramps.flap_factor:initial_val', val=0.)
        self.set_input_defaults('aero_ramps.gear_factor:initial_val', val=0.)



import openmdao.api as om

from aviary.mission.flops_based.phases.simplified_takeoff import TakeoffGroup
from aviary.variable_info.variables import Dynamic


class Takeoff:
    """
    Define user constraints for a climb phase.
    Parameters
    ----------
    airport_altitude : float (None)
        altitude of airport (ft)
        ratio of lift to drag at takeoff (None)
    num_engines int (None)
        number of engines (None)

    Returns
    -------
    Group
        a Group object in OpenMDAO
    """

    def __init__(
        self,
        airport_altitude=None,
        ramp_mass=None,
        num_engines=None,
    ):
        self.airport_altitude = airport_altitude  # ft
        self.num_engines = num_engines
        # note: need to clean this up so that some of these variables come from
        # connections. The only variables that should stay are: airport_altitude

    __slots__ = (
        "airport_altitude",
        "num_engines",
    )

    def build_phase(self, use_detailed=False):
        """
        Construct and return a new phase for takeoff analysis.
        Parameters
        ----------
        use_detailed : bool(False)
            tells whether to use simplified or detailed takeoff. Currently detailed is
            disabled.
        Returns
        -------
        Group
            a group in OpenMDAO
        """

        if use_detailed:
            raise om.AnalysisError(
                "Must set takeoff method to `use_detailed=False`, detailed takeoff is"
                " not currently enabled."
            )

        ##############
        # Add Inputs #
        ##############

        takeoff = TakeoffGroup(num_engines=self.num_engines)
        takeoff.set_input_defaults(
            Dynamic.Mission.ALTITUDE,
            val=self.airport_altitude,
            units="ft")

        return takeoff



'''
Define utilities to prepare derived values of aircraft geometry for
aerodynamics analysis.

TODO: blended-wing-body support
TODO: multiple engine model support
'''
import openmdao.api as om
from numpy import pi

from aviary.subsystems.geometry.flops_based.canard import Canard
from aviary.subsystems.geometry.flops_based.characteristic_lengths import \
    CharacteristicLengths
from aviary.subsystems.geometry.flops_based.fuselage import FuselagePrelim
from aviary.subsystems.geometry.flops_based.nacelle import Nacelles
from aviary.subsystems.geometry.flops_based.utils import (
    Names, calc_fuselage_adjustment, calc_lifting_surface_scaler,
    d_calc_fuselage_adjustment, thickness_to_chord_scaler)
from aviary.subsystems.geometry.flops_based.wetted_area_total import TotalWettedArea
from aviary.subsystems.geometry.flops_based.wing import WingPrelim
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft


class PrepGeom(om.Group):
    '''
    Prepare derived values of aircraft geometry for aerodynamics analysis.
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        aviary_options = self.options['aviary_options']

        self.add_subsystem(
            'fuselage_prelim', FuselagePrelim(aviary_options=aviary_options),
            promotes_inputs=['*'],
            promotes_outputs=['*']
        )

        self.add_subsystem(
            'wing_prelim', WingPrelim(aviary_options=aviary_options),
            promotes_inputs=['*'],
            promotes_outputs=['*']
        )

        self.add_subsystem(
            'prelim', _Prelim(aviary_options=aviary_options),
            promotes_inputs=['*'],
        )

        self.add_subsystem(
            'wing', _Wing(aviary_options=aviary_options),
            promotes_inputs=['aircraft*'],
            promotes_outputs=['*']
        )

        self.connect(f'prelim.{Names.CROOT}', f'wing.{Names.CROOT}')
        self.connect(f'prelim.{Names.CROOTB}', f'wing.{Names.CROOTB}')
        self.connect(f'prelim.{Names.XDX}', f'wing.{Names.XDX}')
        self.connect(f'prelim.{Names.XMULT}', f'wing.{Names.XMULT}')

        self.add_subsystem(
            'tail', _Tail(aviary_options=aviary_options),
            promotes_inputs=['aircraft*'],
            promotes_outputs=['*']
        )

        self.connect(f'prelim.{Names.XMULTH}', f'tail.{Names.XMULTH}')
        self.connect(f'prelim.{Names.XMULTV}', f'tail.{Names.XMULTV}')

        self.add_subsystem(
            'fuselage', _Fuselage(aviary_options=aviary_options),
            promotes_inputs=['aircraft*'],
            promotes_outputs=['*']
        )

        self.connect(f'prelim.{Names.CROOTB}', f'fuselage.{Names.CROOTB}')
        self.connect(f'prelim.{Names.CROTVT}', f'fuselage.{Names.CROTVT}')
        self.connect(f'prelim.{Names.CRTHTB}', f'fuselage.{Names.CRTHTB}')

        self.add_subsystem(
            'nacelles', Nacelles(aviary_options=aviary_options),
            promotes_inputs=['aircraft*'],
            promotes_outputs=['*']
        )

        self.add_subsystem(
            'canard', Canard(aviary_options=aviary_options),
            promotes_inputs=['aircraft*'],
            promotes_outputs=['*']
        )

        self.add_subsystem(
            'characteristic_lengths',
            CharacteristicLengths(aviary_options=aviary_options),
            promotes_inputs=['aircraft*'],
            promotes_outputs=['*']
        )

        self.connect(
            f'prelim.{Names.CROOT}', f'characteristic_lengths.{Names.CROOT}'
        )

        self.add_subsystem(
            'total_wetted_area', TotalWettedArea(aviary_options=aviary_options),
            promotes_inputs=['*'],
            promotes_outputs=['*']
        )


class _Prelim(om.ExplicitComponent):
    '''
    Calculate internal derived values for of aircraft geometry for aerodynamics
    analysis.
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.Fuselage.AVG_DIAMETER, 0.0)
        add_aviary_input(self, Aircraft.Fuselage.MAX_WIDTH, 0.0)

        add_aviary_input(self, Aircraft.HorizontalTail.AREA, 0.0)
        add_aviary_input(self, Aircraft.HorizontalTail.ASPECT_RATIO, 0.0)
        add_aviary_input(self, Aircraft.HorizontalTail.TAPER_RATIO, 0.0)
        add_aviary_input(self, Aircraft.HorizontalTail.THICKNESS_TO_CHORD, 0.0)

        add_aviary_input(self, Aircraft.VerticalTail.AREA, 0.0)
        add_aviary_input(self, Aircraft.VerticalTail.ASPECT_RATIO, 0.0)
        add_aviary_input(self, Aircraft.VerticalTail.TAPER_RATIO, 0.0)
        add_aviary_input(self, Aircraft.VerticalTail.THICKNESS_TO_CHORD, 0.0)

        add_aviary_input(self, Aircraft.Wing.AREA, 0.0)
        add_aviary_input(self, Aircraft.Wing.GLOVE_AND_BAT, 0.0)
        # NOTE: FLOPS/aviary1 calculate span locally
        add_aviary_input(self, Aircraft.Wing.SPAN, 0.0)
        add_aviary_input(self, Aircraft.Wing.TAPER_RATIO, 0.0)
        add_aviary_input(self, Aircraft.Wing.THICKNESS_TO_CHORD, 0.0)

        self.add_output(Names.CROOT, 1.0, units='unitless')
        self.add_output(Names.CROOTB, 1.0, units='unitless')
        self.add_output(Names.CROTM, 1.0, units='unitless')
        self.add_output(Names.CROTVT, 1.0, units='unitless')
        self.add_output(Names.CRTHTB, 1.0, units='unitless')
        self.add_output(Names.SPANHT, 1.0, units='unitless')
        self.add_output(Names.SPANVT, 1.0, units='unitless')
        self.add_output(Names.XDX, 1.0, units='unitless')
        self.add_output(Names.XMULT, 1.0, units='unitless')
        self.add_output(Names.XMULTH, 1.0, units='unitless')
        self.add_output(Names.XMULTV, 1.0, units='unitless')

    def setup_partials(self):
        fuselage_var = self.fuselage_var

        self.declare_partials(Names.XDX, fuselage_var, val=1.)

        self.declare_partials(
            Names.XMULT, Aircraft.Wing.THICKNESS_TO_CHORD,
            val=thickness_to_chord_scaler
        )

        self.declare_partials(
            Names.XMULTH, Aircraft.HorizontalTail.THICKNESS_TO_CHORD,
            val=thickness_to_chord_scaler
        )

        self.declare_partials(
            Names.XMULTV, Aircraft.VerticalTail.THICKNESS_TO_CHORD,
            val=thickness_to_chord_scaler
        )

        self.declare_partials(
            Names.SPANHT,
            [
                Aircraft.HorizontalTail.AREA,
                Aircraft.HorizontalTail.ASPECT_RATIO,
            ]
        )

        self.declare_partials(
            Names.CRTHTB,
            [
                Aircraft.HorizontalTail.AREA,
                Aircraft.HorizontalTail.ASPECT_RATIO,
                Aircraft.HorizontalTail.TAPER_RATIO,
                fuselage_var,
            ]
        )

        self.declare_partials(
            Names.CROOT,
            [
                Aircraft.Wing.AREA,
                Aircraft.Wing.GLOVE_AND_BAT,
                Aircraft.Wing.SPAN,
                Aircraft.Wing.TAPER_RATIO,
            ]
        )

        self.declare_partials(
            Names.CROTM,
            [
                Aircraft.Wing.SPAN,
                Aircraft.Wing.TAPER_RATIO,
                fuselage_var,
            ]
        )

        self.declare_partials(
            Names.CROOTB,
            [
                Aircraft.Wing.AREA,
                Aircraft.Wing.GLOVE_AND_BAT,
                Aircraft.Wing.SPAN,
                Aircraft.Wing.TAPER_RATIO,
                fuselage_var,
            ]
        )

        self.declare_partials(
            Names.SPANVT,
            [
                Aircraft.VerticalTail.AREA,
                Aircraft.VerticalTail.ASPECT_RATIO,
            ]
        )

        self.declare_partials(
            Names.CROTVT,
            [
                Aircraft.VerticalTail.AREA,
                Aircraft.VerticalTail.ASPECT_RATIO,
                Aircraft.VerticalTail.TAPER_RATIO,
            ]
        )

    def compute(
        self, inputs, outputs, discrete_inputs=None, discrete_outputs=None
    ):
        w_tc = inputs[Aircraft.Wing.THICKNESS_TO_CHORD]
        h_tc = inputs[Aircraft.HorizontalTail.THICKNESS_TO_CHORD]
        v_tc = inputs[Aircraft.VerticalTail.THICKNESS_TO_CHORD]

        outputs[Names.XMULT] = calc_lifting_surface_scaler(w_tc)
        outputs[Names.XMULTH] = calc_lifting_surface_scaler(h_tc)
        outputs[Names.XMULTV] = calc_lifting_surface_scaler(v_tc)

        fuselage_var = self.fuselage_var

        XDX = outputs[Names.XDX] = inputs[fuselage_var]
        aspect_ratio = inputs[Aircraft.HorizontalTail.ASPECT_RATIO]
        area = inputs[Aircraft.HorizontalTail.AREA]

        span = outputs[Names.SPANHT] = (aspect_ratio * area)**0.5

        CRTHTB = 0.0

        if 0.0 < span:
            taper_ratio = inputs[Aircraft.HorizontalTail.TAPER_RATIO]

            CRTHTB = (
                2.0 * area / (span * (1.0 + taper_ratio))
                + ((span / 2.0 - XDX / 4.0) / (span / 2.0))
                * (1.0 - taper_ratio) + taper_ratio
            )

        outputs[Names.CRTHTB] = CRTHTB

        area = inputs[Aircraft.Wing.AREA]
        glove_and_bat = inputs[Aircraft.Wing.GLOVE_AND_BAT]
        span = inputs[Aircraft.Wing.SPAN]
        taper_ratio = inputs[Aircraft.Wing.TAPER_RATIO]

        CROOT = outputs[Names.CROOT] = \
            ((area - glove_and_bat) * 2.0) / ((1.0 + taper_ratio) * span)

        CROTM = outputs[Names.CROTM] = (
            ((span / 2.0 - XDX / 2.0) / (span / 2.0)) * (1.0 - taper_ratio)
            + taper_ratio
        )

        outputs[Names.CROOTB] = CROOT * CROTM

        area = inputs[Aircraft.VerticalTail.AREA]
        aspect_ratio = inputs[Aircraft.VerticalTail.ASPECT_RATIO]

        span = outputs[Names.SPANVT] = (area * aspect_ratio)**0.5

        CROTVT = 0.0

        if 0.0 < span:
            taper_ratio = inputs[Aircraft.VerticalTail.TAPER_RATIO]

            CROTVT = 2.0 * area / (span * (1.0 + taper_ratio))

        outputs[Names.CROTVT] = CROTVT

    def compute_partials(self, inputs, J, discrete_inputs=None):
        fuselage_var = self.fuselage_var

        XDX = inputs[fuselage_var]
        area = inputs[Aircraft.HorizontalTail.AREA]
        aspect_ratio = inputs[Aircraft.HorizontalTail.ASPECT_RATIO]

        span2 = area * aspect_ratio
        span = span2**0.5
        f = 0.5 / span

        J[Names.SPANHT, Aircraft.HorizontalTail.AREA] = f * aspect_ratio
        J[Names.SPANHT, Aircraft.HorizontalTail.ASPECT_RATIO] = f * area

        da = dr = dt = dx = 0.0

        if 0.0 < span:
            # b = (a * ar)**0.5
            #
            #        2 * a       b / 2 - x / 4
            # c = ____________ + _____________ * (1 - tr) + tr
            #     b * (1 + tr)       b / 2
            #
            #              2 * a               x * (1 - tr)
            #   = ________________________ - _________________ + 1
            #     (a * ar)**0.5 * (1 + tr)   2 * (a * ar)**0.5
            taper_ratio = inputs[Aircraft.HorizontalTail.TAPER_RATIO]

            _1p_tr = 1.0 + taper_ratio
            _1m_tr = 1.0 - taper_ratio

            # da = d(f0 / g0) + d(f1 / g1) + 0
            #      df0 * g0 - f0 * dg0   df1 * g1 - f1 * dg1
            #    = ___________________ + ___________________
            #             g0**2                 g1**2
            dspan_darea = 0.5 * (aspect_ratio / area) ** 0.5

            da = (
                2.0 / _1p_tr * (1.0 - area * dspan_darea / span) / span
                + _1m_tr * 0.5 * XDX * dspan_darea / span ** 2
            )

            # dr = d(k0 * a / (a * ar)**0.5) - d(k1 / (a * ar)**0.5) + 0
            #    = d((k0 * a - k1) / (a * ar)**0.5)
            #    = -0.5 * (k0 * a - k1) / (a * ar)**1.5 * a
            k0 = 2.0 / _1p_tr
            k1 = XDX * _1m_tr / 2.0
            dr = -0.5 * area * (k0 * area - k1) / span**3.0

            # dt = d(k0 / (1 + tr)) - d(k1 * (1 - tr)) + 0
            #    = -k0 / (1 + tr)**2 + k1
            k0 = 2.0 * area / span
            k1 = XDX / (2.0 * span)
            dt = k1 - k0 / _1p_tr**2.0

            # dx = 0 - d(x * k) + 0
            #    = -k
            dx = -_1m_tr / (2.0 * span)

        J[Names.CRTHTB, Aircraft.HorizontalTail.AREA] = da
        J[Names.CRTHTB, Aircraft.HorizontalTail.ASPECT_RATIO] = dr
        J[Names.CRTHTB, Aircraft.HorizontalTail.TAPER_RATIO] = dt
        J[Names.CRTHTB, fuselage_var] = dx

        area = inputs[Aircraft.Wing.AREA]
        glove_and_bat = inputs[Aircraft.Wing.GLOVE_AND_BAT]
        span = inputs[Aircraft.Wing.SPAN]
        taper_ratio = inputs[Aircraft.Wing.TAPER_RATIO]

        a_g = area - glove_and_bat
        _1p_tr = 1.0 + taper_ratio
        span_1p_tr = span * _1p_tr

        #     (a - g) * 2
        # c = ____________
        #     (1 + tr) * b
        CROOT = 2.0 * a_g / span_1p_tr

        dc = J[Names.CROOT, Aircraft.Wing.AREA] = 2.0 / span_1p_tr
        J[Names.CROOT, Aircraft.Wing.GLOVE_AND_BAT] = -dc
        J[Names.CROOT, Aircraft.Wing.SPAN] = -CROOT / span
        J[Names.CROOT, Aircraft.Wing.TAPER_RATIO] = -CROOT / _1p_tr

        #
        #     (b / 2) - (x / 2)
        # c = _________________ * (1 - tr) + tr
        #          (b / 2)

        # db = d(f / g) + 0
        #    = ((df * g) - (f * dg)) / g**2
        _1m_tr = 1.0 - taper_ratio
        g = span / 2.0
        f = (g - (XDX / 2.0)) * _1m_tr
        df = _1m_tr / 2.0
        dg = 0.5
        J[Names.CROTM, Aircraft.Wing.SPAN] = (df * g - f * dg) / g**2.0

        # dt = d(k * (1 - tr) + tr)
        k = (g - (XDX / 2.0)) / g
        J[Names.CROTM, Aircraft.Wing.TAPER_RATIO] = -k + 1
        # dx = 0 - d(k * x) + 0 = -k
        J[Names.CROTM, fuselage_var] = -_1m_tr / span

        # dc = d(f * g) = df * g + f * dg
        f = CROOT

        g = (
            (1.0 - taper_ratio) * ((span / 2.0) - (XDX / 2.0)) / (span / 2.0)
            + taper_ratio
        )

        df = J[Names.CROOT, Aircraft.Wing.AREA]
        dg = 0.0
        J[Names.CROOTB, Aircraft.Wing.AREA] = df * g

        df = J[Names.CROOT, Aircraft.Wing.GLOVE_AND_BAT]
        dg = 0.0
        J[Names.CROOTB, Aircraft.Wing.GLOVE_AND_BAT] = df * g

        df = J[Names.CROOT, Aircraft.Wing.SPAN]
        dg = J[Names.CROTM, Aircraft.Wing.SPAN]
        J[Names.CROOTB, Aircraft.Wing.SPAN] = df * g + f * dg

        df = J[Names.CROOT, Aircraft.Wing.TAPER_RATIO]
        dg = J[Names.CROTM, Aircraft.Wing.TAPER_RATIO]
        J[Names.CROOTB, Aircraft.Wing.TAPER_RATIO] = df * g + f * dg

        df = 0.0
        dg = J[Names.CROTM, fuselage_var]
        J[Names.CROOTB, fuselage_var] = f * dg

        area = inputs[Aircraft.VerticalTail.AREA]
        aspect_ratio = inputs[Aircraft.VerticalTail.ASPECT_RATIO]

        span = (area * aspect_ratio)**0.5

        J[Names.SPANVT, Aircraft.VerticalTail.AREA] = \
            0.5 * aspect_ratio / span

        J[Names.SPANVT, Aircraft.VerticalTail.ASPECT_RATIO] = \
            0.5 * area / span

        da = dr = dt = 0.0

        if 0.0 < span:
            taper_ratio = inputs[Aircraft.VerticalTail.TAPER_RATIO]

            _1p_tr = (1.0 + taper_ratio)

            CROTVT = 2.0 * area / (span * _1p_tr)

            # da = d(f / g) = (df * g - f * dg) / g**2
            f = 2.0 * area / _1p_tr
            g = span
            df = 2.0 / _1p_tr
            dg = J[Names.SPANVT, Aircraft.VerticalTail.AREA]
            da = (df * g - f * dg) / g**2

            # dr = d(k / (a * ar)**0.5)
            #    = -0.5 * k / (a * ar)**1.5 * a
            dr = -area**2.0 / (_1p_tr * span**3.0)

            # dt = d(k / (1 + tr)) = -k / (1 + tr)**2
            dt = -2.0 * area / (span * _1p_tr**2.0)

        J[Names.CROTVT, Aircraft.VerticalTail.AREA] = da
        J[Names.CROTVT, Aircraft.VerticalTail.ASPECT_RATIO] = dr
        J[Names.CROTVT, Aircraft.VerticalTail.TAPER_RATIO] = dt

    @property
    def fuselage_var(self):
        '''
        Define the variable name associated with XDX.
        '''
        value = Aircraft.Fuselage.AVG_DIAMETER
        aviary_options: AviaryValues = self.options['aviary_options']

        if aviary_options.get_val(Aircraft.Wing.SPAN_EFFICIENCY_REDUCTION):
            value = Aircraft.Fuselage.MAX_WIDTH

        return value


class _Wing(om.ExplicitComponent):
    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        self.add_input(Names.CROOT, 0.0, units='unitless')
        self.add_input(Names.CROOTB, 0.0, units='unitless')
        self.add_input(Names.XDX, 0.0, units='unitless')
        self.add_input(Names.XMULT, 0.0, units='unitless')

        add_aviary_input(self, Aircraft.Wing.AREA, 0.0)
        add_aviary_input(self, Aircraft.Wing.WETTED_AREA_SCALER, 1.0)

        add_aviary_output(self, Aircraft.Wing.WETTED_AREA, 0.0)

    def setup_partials(self):
        self.declare_partials(
            Aircraft.Wing.WETTED_AREA,
            [
                Names.CROOT, Names.CROOTB, Names.XDX, Names.XMULT,
                Aircraft.Wing.AREA, Aircraft.Wing.WETTED_AREA_SCALER,
            ]
        )

    def compute(
        self, inputs, outputs, discrete_inputs=None, discrete_outputs=None
    ):
        aviary_options: AviaryValues = self.options['aviary_options']
        num_fuselage = aviary_options.get_val(Aircraft.Fuselage.NUM_FUSELAGES)

        area = inputs[Aircraft.Wing.AREA]
        CROOT = inputs[Names.CROOT]
        CROOTB = inputs[Names.CROOTB]
        scaler = inputs[Aircraft.Wing.WETTED_AREA_SCALER]
        XDX = inputs[Names.XDX]
        XMULT = inputs[Names.XMULT]

        wetted_area = (
            scaler * XMULT
            * (area - (num_fuselage * XDX / 2.0) * (CROOT + CROOTB))
        )

        outputs[Aircraft.Wing.WETTED_AREA] = wetted_area

    def compute_partials(self, inputs, J, discrete_inputs=None):
        aviary_options: AviaryValues = self.options['aviary_options']
        num_fuselage = aviary_options.get_val(Aircraft.Fuselage.NUM_FUSELAGES)

        area = inputs[Aircraft.Wing.AREA]
        CROOT = inputs[Names.CROOT]
        CROOTB = inputs[Names.CROOTB]
        scaler = inputs[Aircraft.Wing.WETTED_AREA_SCALER]
        XDX = inputs[Names.XDX]
        XMULT = inputs[Names.XMULT]

        J[Aircraft.Wing.WETTED_AREA, Aircraft.Wing.AREA] = scaler * XMULT

        J[Aircraft.Wing.WETTED_AREA, Aircraft.Wing.WETTED_AREA_SCALER] = \
            XMULT * (area - (num_fuselage * XDX / 2.0) * (CROOT + CROOTB))

        J[Aircraft.Wing.WETTED_AREA, Names.CROOT] = \
            - 0.5 * scaler * XMULT * (num_fuselage * XDX)

        J[Aircraft.Wing.WETTED_AREA, Names.CROOTB] = \
            J[Aircraft.Wing.WETTED_AREA, Names.CROOT]

        J[Aircraft.Wing.WETTED_AREA, Names.XDX] = \
            -0.5 * scaler * XMULT * num_fuselage * (CROOT + CROOTB)

        J[Aircraft.Wing.WETTED_AREA, Names.XMULT] = \
            scaler * (area - 0.5*(num_fuselage * XDX) * (CROOT + CROOTB))


class _Tail(om.ExplicitComponent):

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        self.add_input(Names.XMULTH, 0.0, units='unitless')
        self.add_input(Names.XMULTV, 0.0, units='unitless')

        add_aviary_input(self, Aircraft.HorizontalTail.AREA, 0.0)

        add_aviary_input(
            self, Aircraft.HorizontalTail.VERTICAL_TAIL_FRACTION, 0.0
        )

        add_aviary_input(self, Aircraft.HorizontalTail.WETTED_AREA_SCALER, 1.0)

        add_aviary_input(self, Aircraft.VerticalTail.AREA, 0.0)
        add_aviary_input(self, Aircraft.VerticalTail.WETTED_AREA_SCALER, 1.0)

        add_aviary_output(self, Aircraft.HorizontalTail.WETTED_AREA, 0.0)
        add_aviary_output(self, Aircraft.VerticalTail.WETTED_AREA, 0.0)

    def setup_partials(self):
        self.declare_partials(
            Aircraft.HorizontalTail.WETTED_AREA,
            [
                Names.XMULTH, Aircraft.HorizontalTail.AREA,
                Aircraft.HorizontalTail.WETTED_AREA_SCALER
            ]
        )

        self.declare_partials(
            Aircraft.VerticalTail.WETTED_AREA,
            [
                Names.XMULTV, Aircraft.VerticalTail.AREA,
                Aircraft.VerticalTail.WETTED_AREA_SCALER
            ]
        )

        aviary_options: AviaryValues = self.options['aviary_options']
        redux = aviary_options.get_val(Aircraft.Wing.SPAN_EFFICIENCY_REDUCTION)

        if not redux:
            self.declare_partials(
                Aircraft.HorizontalTail.WETTED_AREA,
                [Aircraft.HorizontalTail.VERTICAL_TAIL_FRACTION]
            )

    def compute(
        self, inputs, outputs, discrete_inputs=None, discrete_outputs=None
    ):
        # horizontal tail
        XMULTH = inputs[Names.XMULTH]
        area = inputs[Aircraft.HorizontalTail.AREA]
        scaler = inputs[Aircraft.HorizontalTail.WETTED_AREA_SCALER]

        wetted_area = scaler * XMULTH * area

        aviary_options: AviaryValues = self.options['aviary_options']
        redux = aviary_options.get_val(Aircraft.Wing.SPAN_EFFICIENCY_REDUCTION)

        if not redux:
            num_fuselage_engines = \
                aviary_options.get_val(Aircraft.Propulsion.TOTAL_NUM_FUSELAGE_ENGINES)

            vertical_tail_fraction = \
                inputs[Aircraft.HorizontalTail.VERTICAL_TAIL_FRACTION]

            wetted_area *= (
                1.0 - (0.185 + num_fuselage_engines * 0.063)
                * (1.0 - vertical_tail_fraction)
            )

        outputs[Aircraft.HorizontalTail.WETTED_AREA] = wetted_area

        # vertical tail
        XMULTV = inputs[Names.XMULTV]
        area = inputs[Aircraft.VerticalTail.AREA]
        scaler = inputs[Aircraft.VerticalTail.WETTED_AREA_SCALER]

        wetted_area = scaler * XMULTV * area

        outputs[Aircraft.VerticalTail.WETTED_AREA] = wetted_area

    def compute_partials(self, inputs, J, discrete_inputs=None):
        aviary_options: AviaryValues = self.options['aviary_options']
        redux = aviary_options.get_val(Aircraft.Wing.SPAN_EFFICIENCY_REDUCTION)

        # horizontal tail
        XMULTH = inputs[Names.XMULTH]
        area = inputs[Aircraft.HorizontalTail.AREA]
        scaler = inputs[Aircraft.HorizontalTail.WETTED_AREA_SCALER]

        if not redux:
            num_fuselage_engines = aviary_options.get_val(
                Aircraft.Propulsion.TOTAL_NUM_FUSELAGE_ENGINES)

            vertical_tail_fraction = \
                inputs[Aircraft.HorizontalTail.VERTICAL_TAIL_FRACTION]

            fengines = 0.185 + num_fuselage_engines * 0.063
            fact = 1.0 - fengines * (1.0 - vertical_tail_fraction)

            J[Aircraft.HorizontalTail.WETTED_AREA, Names.XMULTH] = \
                scaler * area * fact

            J[
                Aircraft.HorizontalTail.WETTED_AREA,
                Aircraft.HorizontalTail.AREA
            ] = scaler * XMULTH * fact

            deriv = scaler * XMULTH * area * fengines

            J[
                Aircraft.HorizontalTail.WETTED_AREA,
                Aircraft.HorizontalTail.VERTICAL_TAIL_FRACTION
            ] = deriv

            J[
                Aircraft.HorizontalTail.WETTED_AREA,
                Aircraft.HorizontalTail.WETTED_AREA_SCALER,
            ] = XMULTH * area * fact

        else:
            J[Aircraft.HorizontalTail.WETTED_AREA, Names.XMULTH] = \
                scaler * area

            J[
                Aircraft.HorizontalTail.WETTED_AREA,
                Aircraft.HorizontalTail.AREA
            ] = scaler * XMULTH

            J[
                Aircraft.HorizontalTail.WETTED_AREA,
                Aircraft.HorizontalTail.WETTED_AREA_SCALER,
            ] = XMULTH * area

        # vertical tail
        XMULTV = inputs[Names.XMULTV]
        area = inputs[Aircraft.VerticalTail.AREA]
        scaler = inputs[Aircraft.VerticalTail.WETTED_AREA_SCALER]

        J[Aircraft.VerticalTail.WETTED_AREA, Names.XMULTV] = scaler * area

        J[
            Aircraft.VerticalTail.WETTED_AREA,
            Aircraft.VerticalTail.AREA
        ] = scaler * XMULTV

        J[
            Aircraft.VerticalTail.WETTED_AREA,
            Aircraft.VerticalTail.WETTED_AREA_SCALER
        ] = XMULTV * area


class _Fuselage(om.ExplicitComponent):
    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        self.add_input(Names.CROOTB, 0.0, units='unitless')
        self.add_input(Names.CROTVT, 0.0, units='unitless')
        self.add_input(Names.CRTHTB, 0.0, units='unitless')

        add_aviary_input(self, Aircraft.Fuselage.AVG_DIAMETER, 0.0)
        add_aviary_input(self, Aircraft.Fuselage.LENGTH, 0.0)
        add_aviary_input(self, Aircraft.Fuselage.WETTED_AREA_SCALER, 1.0)

        add_aviary_input(self, Aircraft.HorizontalTail.THICKNESS_TO_CHORD, 0.0)

        add_aviary_input(
            self, Aircraft.HorizontalTail.VERTICAL_TAIL_FRACTION, 0.0
        )

        add_aviary_input(self, Aircraft.VerticalTail.THICKNESS_TO_CHORD, 0.0)

        add_aviary_input(self, Aircraft.Wing.AREA, 0.0)
        add_aviary_input(self, Aircraft.Wing.ASPECT_RATIO, 0.0)
        add_aviary_input(self, Aircraft.Wing.GLOVE_AND_BAT, 0.0)
        add_aviary_input(self, Aircraft.Wing.THICKNESS_TO_CHORD, 0.0)

        add_aviary_output(self, Aircraft.Fuselage.CROSS_SECTION, 0.0)
        add_aviary_output(self, Aircraft.Fuselage.DIAMETER_TO_WING_SPAN, 0.0)
        add_aviary_output(self, Aircraft.Fuselage.LENGTH_TO_DIAMETER, 0.0)
        add_aviary_output(self, Aircraft.Fuselage.WETTED_AREA, 0.0)

    def setup_partials(self):
        self.declare_partials(
            Aircraft.Fuselage.DIAMETER_TO_WING_SPAN,
            [
                Aircraft.Wing.AREA,
                Aircraft.Wing.ASPECT_RATIO,
                Aircraft.Fuselage.AVG_DIAMETER,
                Aircraft.Wing.GLOVE_AND_BAT,
            ]
        )

        self.declare_partials(
            Aircraft.Fuselage.CROSS_SECTION,
            Aircraft.Fuselage.AVG_DIAMETER
        )

        self.declare_partials(
            Aircraft.Fuselage.LENGTH_TO_DIAMETER,
            [
                Aircraft.Fuselage.LENGTH,
                Aircraft.Fuselage.AVG_DIAMETER,
            ]
        )

        self.declare_partials(
            Aircraft.Fuselage.WETTED_AREA,
            [
                Aircraft.Fuselage.AVG_DIAMETER,
                Aircraft.Fuselage.LENGTH,
                Aircraft.Fuselage.WETTED_AREA_SCALER,
                Aircraft.HorizontalTail.THICKNESS_TO_CHORD,
                Aircraft.HorizontalTail.VERTICAL_TAIL_FRACTION,
                Aircraft.VerticalTail.THICKNESS_TO_CHORD,
                Aircraft.Wing.THICKNESS_TO_CHORD,
                Names.CROOTB,
                Names.CROTVT,
                Names.CRTHTB,
            ]
        )

    def compute(
        self, inputs, outputs, discrete_inputs=None, discrete_outputs=None
    ):
        aviary_options: AviaryValues = self.options['aviary_options']
        num_fuselages = aviary_options.get_val(Aircraft.Fuselage.NUM_FUSELAGES)

        area = inputs[Aircraft.Wing.AREA]
        aspect_ratio = inputs[Aircraft.Wing.ASPECT_RATIO]
        avg_diam = inputs[Aircraft.Fuselage.AVG_DIAMETER]
        glove_and_bat = inputs[Aircraft.Wing.GLOVE_AND_BAT]

        diam_to_wing_span = \
            avg_diam / (aspect_ratio * (area - glove_and_bat))**0.5

        outputs[Aircraft.Fuselage.DIAMETER_TO_WING_SPAN] = diam_to_wing_span

        cross_section = pi * (avg_diam / 2.0)**2.0

        outputs[Aircraft.Fuselage.CROSS_SECTION] = cross_section

        length_to_diam = 100.0  # FLOPS default value

        if 0.0 < avg_diam:
            length = inputs[Aircraft.Fuselage.LENGTH]

            length_to_diam = length / avg_diam

        outputs[Aircraft.Fuselage.LENGTH_TO_DIAMETER] = length_to_diam

        wetted_area = 0.0
        scaler = inputs[Aircraft.Fuselage.WETTED_AREA_SCALER]

        if (0 < num_fuselages) and (0.0 < avg_diam):
            CROOTB = inputs[Names.CROOTB]
            thickness_chord = inputs[Aircraft.Wing.THICKNESS_TO_CHORD]

            CRTHTB = inputs[Names.CRTHTB]

            ht_thickness_chord = \
                inputs[Aircraft.HorizontalTail.THICKNESS_TO_CHORD]

            vertical_tail_fraction = \
                inputs[Aircraft.HorizontalTail.VERTICAL_TAIL_FRACTION]

            CROTVT = inputs[Names.CROTVT]

            vt_thickness_chord = \
                inputs[Aircraft.VerticalTail.THICKNESS_TO_CHORD]

            cfa = calc_fuselage_adjustment(CROOTB, thickness_chord)
            cfah = calc_fuselage_adjustment(CRTHTB, ht_thickness_chord)
            cfav = calc_fuselage_adjustment(CROTVT, vt_thickness_chord)

            wetted_area = scaler * (
                pi * avg_diam ** 2.0
                * (length / avg_diam - 1.7) - 2.0 * cfa - 2.0 * cfah
                * (1.0 - vertical_tail_fraction) - cfav
            )

        outputs[Aircraft.Fuselage.WETTED_AREA] = wetted_area

    def compute_partials(self, inputs, J, discrete_inputs=None):
        aviary_options: AviaryValues = self.options['aviary_options']
        num_fuselages = aviary_options.get_val(Aircraft.Fuselage.NUM_FUSELAGES)

        area = inputs[Aircraft.Wing.AREA]
        aspect_ratio = inputs[Aircraft.Wing.ASPECT_RATIO]
        avg_diam = inputs[Aircraft.Fuselage.AVG_DIAMETER]
        glove_and_bat = inputs[Aircraft.Wing.GLOVE_AND_BAT]

        fact = aspect_ratio * (area - glove_and_bat)
        fact2 = 1.0 / fact ** 1.5

        J[
            Aircraft.Fuselage.DIAMETER_TO_WING_SPAN,
            Aircraft.Fuselage.AVG_DIAMETER
        ] = 1.0 / fact ** 0.5

        J[
            Aircraft.Fuselage.DIAMETER_TO_WING_SPAN,
            Aircraft.Wing.ASPECT_RATIO
        ] = -0.5 * avg_diam * (area - glove_and_bat) * fact2

        J[Aircraft.Fuselage.DIAMETER_TO_WING_SPAN, Aircraft.Wing.AREA] = \
            -0.5 * avg_diam * aspect_ratio * fact2

        J[
            Aircraft.Fuselage.DIAMETER_TO_WING_SPAN,
            Aircraft.Wing.GLOVE_AND_BAT
        ] = 0.5 * avg_diam * aspect_ratio * fact2

        J[Aircraft.Fuselage.CROSS_SECTION, Aircraft.Fuselage.AVG_DIAMETER] = \
            0.5 * pi * avg_diam

        if 0.0 < avg_diam:
            length = inputs[Aircraft.Fuselage.LENGTH]

            J[
                Aircraft.Fuselage.LENGTH_TO_DIAMETER,
                Aircraft.Fuselage.AVG_DIAMETER
            ] = -length / avg_diam ** 2

            J[
                Aircraft.Fuselage.LENGTH_TO_DIAMETER, Aircraft.Fuselage.LENGTH
            ] = 1.0 / avg_diam

        else:
            J[
                Aircraft.Fuselage.LENGTH_TO_DIAMETER,
                Aircraft.Fuselage.AVG_DIAMETER
            ] = J[
                Aircraft.Fuselage.LENGTH_TO_DIAMETER, Aircraft.Fuselage.LENGTH
            ] = 0.0

        if (0 < num_fuselages) and (0.0 < avg_diam):
            CROOTB = inputs[Names.CROOTB]
            CRTHTB = inputs[Names.CRTHTB]
            CROTVT = inputs[Names.CROTVT]

            scaler = inputs[Aircraft.Fuselage.WETTED_AREA_SCALER]

            thickness_chord = inputs[Aircraft.Wing.THICKNESS_TO_CHORD]

            ht_thickness_chord = \
                inputs[Aircraft.HorizontalTail.THICKNESS_TO_CHORD]

            vt_thickness_chord = \
                inputs[Aircraft.VerticalTail.THICKNESS_TO_CHORD]

            vertical_tail_fraction = \
                inputs[Aircraft.HorizontalTail.VERTICAL_TAIL_FRACTION]

            cfa = calc_fuselage_adjustment(CROOTB, thickness_chord)
            cfah = calc_fuselage_adjustment(CRTHTB, ht_thickness_chord)
            cfav = calc_fuselage_adjustment(CROTVT, vt_thickness_chord)

            dcfa = d_calc_fuselage_adjustment(CROOTB, thickness_chord)
            dcfah = d_calc_fuselage_adjustment(CRTHTB, ht_thickness_chord)
            dcfav = d_calc_fuselage_adjustment(CROTVT, vt_thickness_chord)

            J[
                Aircraft.Fuselage.WETTED_AREA, Aircraft.Fuselage.AVG_DIAMETER
            ] = scaler * pi * (length - 3.4 * avg_diam)

            J[Aircraft.Fuselage.WETTED_AREA, Aircraft.Fuselage.LENGTH] = \
                scaler * pi * avg_diam

            J[
                Aircraft.Fuselage.WETTED_AREA,
                Aircraft.Fuselage.WETTED_AREA_SCALER
            ] = (
                pi * avg_diam ** 2.0
                * (length / avg_diam - 1.7) - 2.0 * cfa - 2.0 * cfah
                * (1.0 - vertical_tail_fraction) - cfav
            )

            J[Aircraft.Fuselage.WETTED_AREA, Names.CROOTB] = \
                scaler * -2.0 * dcfa[0]

            J[Aircraft.Fuselage.WETTED_AREA, Names.CRTHTB] = \
                scaler * -2.0 * dcfah[0] * (1.0 - vertical_tail_fraction)

            J[Aircraft.Fuselage.WETTED_AREA, Names.CROTVT] = \
                scaler * -dcfav[0]

            J[
                Aircraft.Fuselage.WETTED_AREA, Aircraft.Wing.THICKNESS_TO_CHORD
            ] = scaler * -2.0 * dcfa[1]

            J[
                Aircraft.Fuselage.WETTED_AREA,
                Aircraft.HorizontalTail.THICKNESS_TO_CHORD
            ] = scaler * -2.0 * dcfah[1] * (1.0 - vertical_tail_fraction)

            J[
                Aircraft.Fuselage.WETTED_AREA,
                Aircraft.VerticalTail.THICKNESS_TO_CHORD
            ] = scaler * -dcfav[1]

            J[
                Aircraft.Fuselage.WETTED_AREA,
                Aircraft.HorizontalTail.VERTICAL_TAIL_FRACTION
            ] = scaler * 2.0 * cfah

        else:
            J[
                Aircraft.Fuselage.WETTED_AREA, Names.CROOTB
            ] = J[
                Aircraft.Fuselage.WETTED_AREA, Names.CRTHTB
            ] = J[
                Aircraft.Fuselage.WETTED_AREA, Names.CROTVT
            ] = J[
                Aircraft.Fuselage.WETTED_AREA, Aircraft.Fuselage.AVG_DIAMETER
            ] = J[
                Aircraft.Fuselage.WETTED_AREA, Aircraft.Fuselage.LENGTH
            ] = J[
                Aircraft.Fuselage.WETTED_AREA,
                Aircraft.Fuselage.WETTED_AREA_SCALER
            ] = J[
                Aircraft.Fuselage.WETTED_AREA,
                Aircraft.HorizontalTail.THICKNESS_TO_CHORD
            ] = J[
                Aircraft.Fuselage.WETTED_AREA,
                Aircraft.HorizontalTail.VERTICAL_TAIL_FRACTION
            ] = J[
                Aircraft.Fuselage.WETTED_AREA,
                Aircraft.VerticalTail.THICKNESS_TO_CHORD
            ] = J[
                Aircraft.Fuselage.WETTED_AREA, Aircraft.Wing.THICKNESS_TO_CHORD
            ] = 0.0



import numpy as np
import openmdao.api as om
from openmdao.utils.units import convert_units

from aviary.constants import GRAV_METRIC_GASP, GRAV_ENGLISH_LBM, MU_TAKEOFF
from aviary.variable_info.functions import add_aviary_input
from aviary.variable_info.variables import Aircraft, Dynamic

LBF_TO_N = convert_units(1.0, 'lbf', 'N')


class UnsteadySolvedEOM(om.ExplicitComponent):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def initialize(self):
        self.options.declare("num_nodes", types=int)
        self.options.declare("ground_roll", types=bool, default=False,
                             desc="True if the aircraft is confined to the ground. "
                                  "Removes altitude rate as an output and adjust "
                                  "the TAS rate equation.")

    def setup(self):
        nn = self.options["num_nodes"]

        # Inputs

        self.add_input("TAS", shape=nn,
                       desc="true air speed", units="m/s")

        # TODO: This should probably be declared in Newtons, but the weight variable
        # is really a mass. This should be resolved with an adapter component that
        # uses gravity.
        self.add_input("mass", shape=nn, desc="aircraft mass", units="lbm")
        self.add_input(Dynamic.Mission.THRUST_TOTAL, shape=nn,
                       desc=Dynamic.Mission.THRUST_TOTAL, units="N")
        self.add_input(Dynamic.Mission.LIFT, shape=nn,
                       desc=Dynamic.Mission.LIFT, units="N")
        self.add_input(Dynamic.Mission.DRAG, shape=nn,
                       desc=Dynamic.Mission.DRAG, units="N")
        add_aviary_input(self, Aircraft.Wing.INCIDENCE, val=0, units="rad")
        self.add_input("alpha", val=np.zeros(
            nn), desc="angle of attack", units="rad")

        if not self.options["ground_roll"]:
            self.add_input(Dynamic.Mission.FLIGHT_PATH_ANGLE, val=np.zeros(
                nn), desc="flight path angle", units="rad")
            self.add_input("dh_dr", val=np.zeros(
                nn), desc="d(alt)/d(range)", units="m/distance_units")
            self.add_input("d2h_dr2", val=np.zeros(
                nn), desc="d(climb_rate)/d(range)", units="1/distance_units")

        # Outputs
        self.add_output("dt_dr", shape=nn, units="s/distance_units",
                        desc="Seconds passed per each meter of range covered.",
                        tags=['dymos.state_rate_source:time', 'dymos.state_units:s'])
        self.add_output("normal_force", val=np.ones(nn), units="N",
                        desc="normal forces")
        self.add_output("fuselage_pitch", val=np.ones(nn), units="rad",
                        desc="fuselage pitch angle")
        self.add_output("load_factor", val=np.ones(nn), units="unitless",
                        desc="load factor")
        self.add_output("dTAS_dt", shape=nn, units="m/s**2",
                        desc="rate of change of true airspeed")

        if not self.options["ground_roll"]:
            self.add_output("dgam_dt", shape=nn, units="rad/s",
                            desc="rate of change of flight path angle", )
            self.add_output("dgam_dt_approx", shape=nn, units="rad/s",
                            desc="approximate rate of change of flight path angle "
                                 "based on dh_dr and d2h_dr2")

    def setup_partials(self):
        nn = self.options["num_nodes"]
        ar = np.arange(nn, dtype=int)
        ground_roll = self.options["ground_roll"]

        self.declare_partials(of="dt_dr", wrt="TAS",
                              rows=ar, cols=ar)

        self.declare_partials(of=["normal_force", "dTAS_dt"],
                              wrt=[Dynamic.Mission.THRUST_TOTAL, Dynamic.Mission.DRAG,
                                   "mass", Dynamic.Mission.LIFT],
                              rows=ar, cols=ar)

        self.declare_partials(of="normal_force", wrt="mass",
                              rows=ar, cols=ar, val=LBF_TO_N * GRAV_ENGLISH_LBM)

        self.declare_partials(of="normal_force", wrt=Dynamic.Mission.LIFT,
                              rows=ar, cols=ar, val=-1.0)

        self.declare_partials(of="load_factor", wrt=[Dynamic.Mission.LIFT, "mass", Dynamic.Mission.THRUST_TOTAL],
                              rows=ar, cols=ar)

        self.declare_partials(of=["dTAS_dt", "normal_force", "load_factor"],
                              wrt=[Aircraft.Wing.INCIDENCE])

        self.declare_partials(of=["normal_force", "dTAS_dt"],
                              wrt=["alpha"],
                              rows=ar, cols=ar)

        self.declare_partials(of="fuselage_pitch",
                              wrt=["alpha"],
                              rows=ar, cols=ar, val=1.0)

        self.declare_partials(of="fuselage_pitch",
                              wrt=Aircraft.Wing.INCIDENCE,
                              val=-1.0)

        self.declare_partials(of="load_factor", wrt=["alpha"],
                              rows=ar, cols=ar)

        if not ground_roll:
            self.declare_partials(of="dt_dr", wrt=Dynamic.Mission.FLIGHT_PATH_ANGLE,
                                  rows=ar, cols=ar)

            self.declare_partials(of=["dgam_dt", "dgam_dt_approx"],
                                  wrt=[Dynamic.Mission.LIFT, "mass", Dynamic.Mission.THRUST_TOTAL,
                                       Dynamic.Mission.DRAG, "alpha", Dynamic.Mission.FLIGHT_PATH_ANGLE],
                                  rows=ar, cols=ar)

            self.declare_partials(of=["normal_force", "dTAS_dt"],
                                  wrt=[Dynamic.Mission.FLIGHT_PATH_ANGLE],
                                  rows=ar, cols=ar)

            self.declare_partials(of=["dgam_dt"], wrt=[
                                  "TAS"], rows=ar, cols=ar)

            self.declare_partials(of="load_factor", wrt=[Dynamic.Mission.FLIGHT_PATH_ANGLE],
                                  rows=ar, cols=ar)

            self.declare_partials(of=["dgam_dt", "dgam_dt_approx"],
                                  wrt=[Dynamic.Mission.LIFT, "mass",
                                       Dynamic.Mission.THRUST_TOTAL, "alpha", Dynamic.Mission.FLIGHT_PATH_ANGLE],
                                  rows=ar, cols=ar)

            self.declare_partials(of="fuselage_pitch",
                                  wrt=[Dynamic.Mission.FLIGHT_PATH_ANGLE],
                                  rows=ar, cols=ar, val=1.0)

            self.declare_partials(of=["dgam_dt_approx"],
                                  wrt=["dh_dr", "d2h_dr2", "TAS"],
                                  rows=ar, cols=ar)

            self.declare_partials(of=["dgam_dt_approx", "dgam_dt"],
                                  wrt=[Aircraft.Wing.INCIDENCE])

    def compute(self, inputs, outputs):
        tas = inputs["TAS"]
        thrust = inputs[Dynamic.Mission.THRUST_TOTAL]
        # convert to newtons  # TODO: change this to use the units conversion
        weight = inputs["mass"] * GRAV_ENGLISH_LBM * LBF_TO_N
        drag = inputs[Dynamic.Mission.DRAG]
        lift = inputs[Dynamic.Mission.LIFT]
        alpha = inputs["alpha"]

        i_wing = inputs[Aircraft.Wing.INCIDENCE]

        g = GRAV_METRIC_GASP
        m = weight / g

        if self.options["ground_roll"]:
            mu = MU_TAKEOFF
            gamma = 0.0
        else:
            mu = 0.0
            gamma = inputs[Dynamic.Mission.FLIGHT_PATH_ANGLE]
            dh_dr = inputs["dh_dr"]
            d2h_dr2 = inputs["d2h_dr2"]

        alpha_i = alpha - i_wing

        cgam = np.cos(gamma)
        sgam = np.sin(gamma)

        calpha_i = np.cos(alpha_i)
        salpha_i = np.sin(alpha_i)

        tcai = thrust * calpha_i
        tsai = thrust * salpha_i

        dr_dt = tas * cgam

        normal_force = weight - lift - tsai

        load_factor = (lift + tsai) / (weight * cgam)

        outputs["dt_dr"] = 1.0 / dr_dt

        outputs["normal_force"] = normal_force

        outputs["dTAS_dt"] = (tcai - drag - weight *
                              sgam - mu * normal_force) / m

        outputs["fuselage_pitch"] = gamma - i_wing + alpha

        outputs["load_factor"] = load_factor

        if not self.options["ground_roll"]:
            outputs["dgam_dt"] = (tsai + lift - weight * cgam) / (m * tas)
            dgam_dr = d2h_dr2 / (dh_dr ** 2 + 1)
            outputs["dgam_dt_approx"] = dgam_dr * dr_dt

    def compute_partials(self, inputs, partials):
        ground_roll = self.options["ground_roll"]

        thrust = inputs[Dynamic.Mission.THRUST_TOTAL]
        # convert to newtons  # TODO: change this to use the units conversion
        weight = inputs["mass"] * GRAV_ENGLISH_LBM * LBF_TO_N
        drag = inputs[Dynamic.Mission.DRAG]
        lift = inputs[Dynamic.Mission.LIFT]
        tas = inputs["TAS"]
        i_wing = inputs[Aircraft.Wing.INCIDENCE]
        alpha = inputs["alpha"]

        if self.options["ground_roll"]:
            mu = MU_TAKEOFF
            gamma = 0.0
        else:
            mu = 0.0
            gamma = inputs[Dynamic.Mission.FLIGHT_PATH_ANGLE]
            dh_dr = inputs["dh_dr"]
            d2h_dr2 = inputs["d2h_dr2"]

        alpha_i = alpha - i_wing

        g = GRAV_METRIC_GASP
        m = weight / g

        mtas = m * tas
        mtas2 = mtas * mtas

        cgam = np.cos(gamma)
        sgam = np.sin(gamma)

        calpha_i = np.cos(alpha_i)
        salpha_i = np.sin(alpha_i)

        dr_dt = tas * cgam

        drdot_dtas = cgam
        drdot_dgam = -tas * sgam

        tcai = thrust * calpha_i
        tsai = thrust * salpha_i

        _f = tcai - drag - weight * sgam - mu * (weight - lift - tsai)

        partials["dt_dr", "TAS"] = -cgam / dr_dt**2

        partials["dTAS_dt", Dynamic.Mission.THRUST_TOTAL] = calpha_i / \
            m + salpha_i / m * mu
        partials["dTAS_dt", Dynamic.Mission.DRAG] = -1. / m

        partials["dTAS_dt", "mass"] = \
            GRAV_ENGLISH_LBM * (LBF_TO_N * (-sgam - mu) / m - _f / (weight/LBF_TO_N * m))

        partials["dTAS_dt", Dynamic.Mission.LIFT] = mu / m
        partials["dTAS_dt", "alpha"] = -tsai / m + mu * tcai / m
        partials["dTAS_dt", Aircraft.Wing.INCIDENCE] = tsai / m - mu * tcai / m

        partials["normal_force", Dynamic.Mission.THRUST_TOTAL] = -salpha_i

        partials["load_factor", Dynamic.Mission.LIFT] = 1 / (weight * cgam)
        partials["load_factor", Dynamic.Mission.THRUST_TOTAL] = salpha_i / \
            (weight * cgam)
        partials["load_factor", "mass"] = \
            - (lift + tsai) / (weight**2/LBF_TO_N * cgam) * GRAV_ENGLISH_LBM

        partials["normal_force", "alpha"] = -tcai
        partials["normal_force", Aircraft.Wing.INCIDENCE] = tcai
        partials["load_factor", Aircraft.Wing.INCIDENCE] = -tcai / (weight * cgam)

        partials["load_factor", "alpha"] = tcai / (weight * cgam)

        if not ground_roll:
            partials["dt_dr", Dynamic.Mission.FLIGHT_PATH_ANGLE] = -drdot_dgam / dr_dt**2

            partials["dTAS_dt", Dynamic.Mission.FLIGHT_PATH_ANGLE] = -weight * cgam / m

            partials["dgam_dt", Dynamic.Mission.THRUST_TOTAL] = salpha_i / mtas
            partials["dgam_dt", Dynamic.Mission.LIFT] = 1. / mtas
            partials["dgam_dt", "mass"] = \
                GRAV_ENGLISH_LBM * (LBF_TO_N*cgam / (mtas) - (tsai +
                                    lift + weight*cgam)/(weight**2 / LBF_TO_N/g * tas))
            partials["dgam_dt", Dynamic.Mission.FLIGHT_PATH_ANGLE] = m * \
                tas * weight * sgam / mtas2
            partials["dgam_dt", "alpha"] = m * tas * tcai / mtas2
            partials["dgam_dt", "TAS"] = - \
                m * (tsai + lift - weight * cgam) / mtas2
            partials["dgam_dt", Aircraft.Wing.INCIDENCE] = -m * tas * tcai / mtas2

            dgam_dr = d2h_dr2 / (dh_dr ** 2 + 1)
            ddgam_dr_ddh_dr = -2 * dh_dr * d2h_dr2 / (dh_dr**2 + 1)**2
            ddgam_dr_dd2h_dr2 = 1. / (dh_dr ** 2 + 1)

            partials["dgam_dt_approx", "dh_dr"] = dr_dt * ddgam_dr_ddh_dr
            partials["dgam_dt_approx", "d2h_dr2"] = dr_dt * ddgam_dr_dd2h_dr2
            partials["dgam_dt_approx", "TAS"] = dgam_dr * drdot_dtas
            partials["dgam_dt_approx",
                     Dynamic.Mission.FLIGHT_PATH_ANGLE] = dgam_dr * drdot_dgam
            partials["load_factor", Dynamic.Mission.FLIGHT_PATH_ANGLE] = (
                lift + tsai) / (weight * cgam**2) * sgam



import warnings
import numpy as np
import openmdao.api as om

from pathlib import Path

from aviary.utils.named_values import get_keys, get_items
from aviary.utils.csv_data_file import read_data_file
from aviary.utils.functions import get_path
from aviary.utils.named_values import NamedValues


def build_data_interpolator(num_nodes, interpolator_data=None, interpolator_outputs=None,
                            method='slinear', extrapolate=True, structured=None,
                            training_data=False):
    """
    Builder for openMDAO metamodel components using data provided via data file, directly
    provided as an argument, or training data passed through openMDAO connections.
    If using a structured grid, data can either be converted from a semistructured
    grid format, or directly provided in structured grid format.

    Parameters
    ----------

    num_nodes : int
        Number of points that will be simultaneously interpolated during model executuion.

    interpolator_data : (str, Path, NamedValues)
        Path to the Aviary csv file containing all data required for interpolation, or
        the data directly given as a NamedValues object.

    interpolator_outputs : dict
        Dictionary describing the names of dependent variables (keys) and their
        units (values). If training_data is False, these variable names must reference
        variables in data_file or interpolator_data. If training_data is True, then
        this dictionary describes the names and units for training data that will be
        provided via openMDAO connections during model execution.

    method : str, optional
        Interpolation method for metamodel. See openMDAO documentation for valid
        options.

    extrapolate : bool, optional
        Flag that sets if metamodel should allow extrapolation

    structured : bool, optional
        Flag to set if interpolation data is a structure grid. If True, the 
        structured metamodel component is used, if False, the semistructured metamodel is
        used. If None, the builder chooses based on provided data structure.

    training_data : bool, optional
        Flag that sets if dependent data for interpolation will be passed via openMDAO
        connections. If True, any provided values for dependent variables will
        be ignored.

    Returns
    -------

    interp_comp : om.MetaModelSemiStructuredComp, om.MetaModelStructuredComp
        OpenMDAO metamodel component using the provided data and flags
    """
    # Argument checking #
    if interpolator_outputs is None:
        raise UserWarning('Independent variables for interpolation were not provided.')
    # if interpolator data is a filepath, get data from file
    if isinstance(interpolator_data, str):
        interpolator_data = get_path(interpolator_data)
    if isinstance(interpolator_data, Path):
        interpolator_data = read_data_file(interpolator_data)

    # Pre-format data: Independent variables placed before dependent variables - position
    #                  of these variables relative to others of their type is preserved
    #                  All data converted to numpy arrays
    indep_vars = NamedValues()
    dep_vars = NamedValues()
    for (key, (val, units)) in get_items(interpolator_data):
        if not isinstance(val, np.ndarray):
            val = np.array(val)
        if key in interpolator_outputs:
            dep_vars.set_val(key, val, units)
        else:
            indep_vars.set_val(key, val, units)
    # update interpolator_data with correctly ordered indep/dep vars in numpy arrays
    interpolator_data.update(indep_vars)
    for (key, (val, units)) in get_items(dep_vars):
        interpolator_data.set_val(key, val, units)

    # TODO investigate creating structured grid from semistructured grid via extrapolation

    # is data already in structured format?
    # assume data is structured until proven otherwise
    data_pre_structured = True
    shape = []
    # check inputs, should be vector of unique values only
    for (key, (val, units)) in get_items(interpolator_data):
        if len(val.shape) == 1:
            if key not in interpolator_outputs:
                # try:
                if np.array_equal(np.unique(val), val):
                    # if vector is only unique values, could be structured!
                    # Store shape and keep going
                    shape.append(len(np.unique(val)))
                else:
                    # Data is not structured. Stop looping through inputs
                    data_pre_structured = False
                    break

    # check outputs, should be array matching shape of input vector lengths
    # if we already know data needs formatting, don't bother checking outputs
    if data_pre_structured:
        for key in interpolator_outputs:
            (val, units) = interpolator_data.get_item(key)
            if np.shape(val) != tuple(shape):
                if len(np.shape(val)) > 1:
                    # we know for sure user was *trying* to set up a structured grid
                    # if output is multi-dimensional array. If output is 1d it could
                    # be a strucured grid with one input, or a semistructured grid
                    raise ValueError(f'shape of output <{key}>, {np.shape(val)}, does '
                                     f'not match expected shape {tuple(shape)}')
                else:
                    # We don't know if data is structured or not if 1d. No harm
                    # in sorting and "reformatting", so assume it needs to be converted
                    data_pre_structured = False
                    break

    if structured is None and data_pre_structured:
        # If the data is already structured, just use a structured grid - it's faster
        # with no downsides
        structured = True
    elif structured is None:
        # In case structured is still None, set it to False - we know data is unstructured
        structured = False

    if not training_data:
        # Sort and format data. Only if not using training data - since we have control
        # over both input and output data they are guarenteed to match after reformatting

        # sort data in semistructured grid format
        # always sort unless data is in structured format
        if not data_pre_structured:
            # first check that data are all vectors of the same length
            for idx, item in enumerate(get_items(interpolator_data)):
                key = item[0]
                units = item[1][1]
                if idx != 0:
                    prev_model_length = model_length
                else:
                    prev_model_length = len(interpolator_data.get_val(key, units))
                model_length = len(interpolator_data.get_val(key, units))
                if model_length != prev_model_length:
                    raise IndexError('Lengths of data provided for interpolation do not '
                                     'match.')

            # get data into column array format
            sorted_values = np.array([val for (key, (val, units))
                                      in get_items(interpolator_data)]).transpose()

            # get all the independent values in format needed for sorting
            independent_vals = np.array([val for (key, (val, units))
                                         in get_items(indep_vars)])

            # Sort by dependent variables in priority order of their appearance
            sorted_values = sorted_values[np.lexsort(np.flip(independent_vals, 0))]

            # reset interpolator_data with sorted values
            for idx, (var, (val, units)) in enumerate(get_items(interpolator_data)):
                interpolator_data.set_val(var, sorted_values[:, idx], units)

        # If user wants structured data, but provided data is not formatted correctly,
        # convert it!
        if structured and not data_pre_structured:
            # Use assumptions for structured grid to format data
            # Only need to reformat data when not using training data, user is responsible
            # for formatting in that case
            # Assumes independent variables are first columns

            (length, var_count) = np.shape(sorted_values)
            indep_var_count = np.shape(independent_vals)[0]

            structured_data = []
            # only need unique independent variables
            unique_data = []
            for i in range(indep_var_count):
                unique_data.append(np.unique(sorted_values[:, i]))
                structured_data.append(unique_data[i])

            shape = tuple([np.size(unique_data[i]) for i in range(indep_var_count)])

            # output data needs to be in nd array format
            for i in range(indep_var_count, var_count):
                structured_data.append(np.reshape(sorted_values[:, i], shape))

            # reset interpolator_data with structured grid formatted values
            for idx, (var, (val, units)) in enumerate(get_items(interpolator_data)):
                interpolator_data.set_val(var, structured_data[idx], units)

    if training_data and structured and not data_pre_structured:
        # User has asked for structured data but not provided it. Use of training data
        # means we can't do any processing on the data including ensuring sorted order,
        # since that might misalign inputs with future connections we can't control here
        # Just convert inputs to structure grid format
        for key in get_keys(indep_vars):
            (val, units) = interpolator_data.get_item(key)
            # take unique values only, put back into interpolator_data
            val = np.unique(val)
            interpolator_data.set_val(key, val, units)

    # create interpolation component
    if structured:
        interp_comp = om.MetaModelStructuredComp(method=method,
                                                 extrapolate=extrapolate,
                                                 vec_size=num_nodes,
                                                 training_data_gradients=training_data)
    else:
        interp_comp = om.MetaModelSemiStructuredComp(method=method,
                                                     extrapolate=extrapolate,
                                                     vec_size=num_nodes,
                                                     training_data_gradients=training_data)

    # add interpolator inputs
    for key in get_keys(indep_vars):
        values, units = interpolator_data.get_item(key)
        interp_comp.add_input(key,
                              training_data=values,
                              units=units)
    # add interpolator outputs
    for key in interpolator_outputs:
        if key in interpolator_data:
            values, units = interpolator_data.get_item(key)
        if training_data:
            units = interpolator_outputs[key]
            interp_comp.add_output(key,
                                   units=units)
        else:
            interp_comp.add_output(key,
                                   training_data=values,
                                   units=units)

    return interp_comp



from copy import deepcopy
from pathlib import Path
import unittest
import warnings

from openmdao.utils.testing_utils import use_tempdirs

from aviary.interface.default_phase_info.height_energy import phase_info
from aviary.interface.methods_for_level1 import run_aviary
from openmdao.utils.testing_utils import use_tempdirs


@unittest.skip("Skipping test due to sensitivity in setup. Need tolerances on actual values")
@use_tempdirs
class BasicReportTestCase(unittest.TestCase):
    def setUp(self):
        local_phase_info = deepcopy(phase_info)
        self.prob = run_aviary('models/test_aircraft/aircraft_for_bench_FwFm.csv',
                               local_phase_info,
                               mission_method="FLOPS", mass_method="FLOPS", optimizer='IPOPT')

    def test_text_report(self):
        self.prob.write_report('test_output.txt')

        expected_text = ['',
                         '',
                         'PROPULSION',
                         'aircraft:engine:scale_factor, 1.0 unitless',
                         'aircraft:propulsion:total_scaled_sls_thrust, 57856.2 lbf',
                         '',
                         'GEOMETRY: FLOPS METHOD',
                         'aircraft:wing:area, 1370.0 ft**2',
                         'aircraft:wing:span, 117.83 ft',
                         'aircraft:wing:aspect_ratio, 11.22091 unitless',
                         'aircraft:wing:sweep, 25.0 deg',
                         'aircraft:horizontal_tail:area, 355.0 ft**2',
                         'aircraft:vertical_tail:area, 284.0 ft**2',
                         'aircraft:fuselage:length, 128.0 ft',
                         'aircraft:fuselage:avg_diameter, 12.75 ft',
                         '',
                         'MASS ESTIMATION: FLOPS REGRESSIONS',
                         'aircraft:wing:mass, 17758.544900084722 lbm',
                         'aircraft:horizontal_tail:mass, 1817.6584389129562 lbm',
                         'aircraft:vertical_tail:mass, 1208.6209771133388 lbm',
                         'aircraft:fins:mass, 0.0 lbm',
                         'aircraft:canard:mass, 0.0 lbm',
                         'aircraft:fuselage:mass, 18357.133455139123 lbm',
                         'aircraft:landing_gear:total_mass, 0.0 lbm',
                         'aircraft:nacelle:mass, 1971.3819954084274 lbm',
                         'aircraft:design:structure_mass, 50200.36816077931 lbm',
                         'aircraft:propulsion:total_engine_mass, 14800.0 lbm',
                         'aircraft:propulsion:total_thrust_reversers_mass, 0.0 lbm',
                         'aircraft:propulsion:total_misc_mass, 648.8373506955577 lbm',
                         'aircraft:fuel:fuel_system_mass, 669.5772386254444 lbm',
                         'aircraft:propulsion:mass, 16118.414589321002 lbm',
                         'aircraft:controls:total_mass, 0.0 lbm',
                         'aircraft:apu:mass, 1142.065041407156 lbm',
                         'aircraft:instruments:mass, 601.1649288403382 lbm',
                         'aircraft:hydraulics:mass, 1086.6955064058336 lbm',
                         'aircraft:electrical:mass, 2463.8713804682375 lbm',
                         'aircraft:avionics:mass, 1652.6486922143051 lbm',
                         'aircraft:furnishings:mass, 15517.315000000002 lbm',
                         'aircraft:air_conditioning:mass, 1601.886853976589 lbm',
                         'aircraft:anti_icing:mass, 208.8500201913504 lbm',
                         'aircraft:design:systems_equip_mass, 25158.285516432577 lbm',
                         'aircraft:design:external_subsystems_mass, 0.0 lbm',
                         'aircraft:design:empty_mass, 91477.0682665329 lbm',
                         'aircraft:crew_and_payload:flight_crew_mass, 450.0 lbm',
                         'aircraft:crew_and_payload:non_flight_crew_mass, 465.0 lbm',
                         'aircraft:fuel:unusable_fuel_mass, 501.30242136015147 lbm',
                         'aircraft:propulsion:total_engine_oil_mass, 130.22722598615468 lbm',
                         'aircraft:crew_and_payload:passenger_service_mass, 3022.748058091815 lbm',
                         'aircraft:crew_and_payload:cargo_container_mass, 1400.00000001059 lbm',
                         'aircraft:design:operating_mass, 97446.3459719816 lbm',
                         'aircraft:crew_and_payload:passenger_mass, 30420.0 lbm',
                         'aircraft:crew_and_payload:passenger_payload_mass, 37856.0 lbm',
                         'aircraft:crew_and_payload:cargo_mass, 0.0 lbm',
                         'aircraft:design:zero_fuel_mass, 135302.3459719816 lbm',
                         'mission:design:fuel_mass, 39469.534814306564 lbm',
                         'mission:summary:total_fuel_mass, 39469.53476849914 lbm',
                         'mission:summary:gross_mass, 0.0 lbm',
                         '',
                         '']

        with open('test_output.txt',
                  'r') as f:
            text = f.readlines()

        for index, line in enumerate(text):
            if index < 1:
                # skip first lines (timestamp)
                continue

            expected_line = ''.join(expected_text[index].split())
            line_no_whitespace = ''.join(line.split())

            # Assert that the lines are equal
            try:
                self.assertEqual(line_no_whitespace.count(expected_line), 1)
            except Exception as error:
                exc_string = f'\nFound:    {line}\nExpected: {expected_text[index]}'
                raise Exception(exc_string)

    def test_file_rename(self):
        # ensure file exists
        with open(Path.cwd() / 'test_output.txt', 'w') as f:
            pass

        # catch warnings as errors
        warnings.filterwarnings("error")

        try:
            self.prob.write_report('test_output.txt')
        except UserWarning:
            pass
        else:
            self.fail('No warning that file output is moved to new name')

        # disable warnings as errors behavior for future tests
        warnings.resetwarnings()


if __name__ == "__main__":
    unittest.main()



import unittest
import os
import shutil
from pathlib import Path

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal

from aviary.utils.functions import add_opts2vals, create_opts2vals
from aviary.variable_info.options import get_option_defaults
from aviary.variable_info.variables import Aircraft, Mission
from aviary.utils.functions import get_path


class TestOpts2Vals(unittest.TestCase):
    def setUp(self):
        self.options = get_option_defaults()
        self.options.set_val(Aircraft.CrewPayload.NUM_PASSENGERS,
                             val=180, units='unitless')
        self.options.set_val(Mission.Design.CRUISE_ALTITUDE, val=35, units='kft')

    def test_default_units(self):
        tol = 1e-4
        self.prob = om.Problem()
        OptionsToValues = create_opts2vals(
            [Aircraft.CrewPayload.NUM_PASSENGERS,
             Mission.Design.CRUISE_ALTITUDE,])
        add_opts2vals(self.prob.model, OptionsToValues, self.options)
        self.prob.setup()
        self.prob.run_model()

        assert_near_equal(
            self.prob['option:'+Aircraft.CrewPayload.NUM_PASSENGERS], 180, tol)
        assert_near_equal(self.prob['option:'+Mission.Design.CRUISE_ALTITUDE], 35, tol)

    def test_specified_units(self):
        tol = 1e-4
        self.prob = om.Problem()
        OptionsToValues = create_opts2vals(
            [Aircraft.CrewPayload.NUM_PASSENGERS,
             Mission.Design.CRUISE_ALTITUDE,],
            output_units={Mission.Design.CRUISE_ALTITUDE: 'm'})
        add_opts2vals(self.prob.model, OptionsToValues, self.options)
        self.prob.setup()
        self.prob.run_model()

        altitude_in_meters = om.convert_units(35000, 'ft', 'm')
        assert_near_equal(
            self.prob['option:'+Aircraft.CrewPayload.NUM_PASSENGERS], 180, tol)
        assert_near_equal(
            self.prob['option:'+Mission.Design.CRUISE_ALTITUDE], altitude_in_meters, tol)


class TestGetPath(unittest.TestCase):

    def setUp(self):
        self.current_dir = Path.cwd()
        process_id = os.getpid()
        # Create a test file in current directory
        self.relative_test_file = f'testfile_{process_id}.txt'
        with open(self.relative_test_file, 'w') as f:
            f.write('test')
        # Create a test file in a sub-directory
        self.sub_directory = self.current_dir / f'tmp_dir_{process_id}'
        self.sub_directory.mkdir(exist_ok=True)
        self.absolute_test_file = self.sub_directory / f'testfile2_{process_id}.txt'
        with open(self.absolute_test_file, 'w') as f:
            f.write('test')

    def tearDown(self):
        # Cleanup the created test files and directories after each test
        if Path(self.relative_test_file).exists():
            Path(self.relative_test_file).unlink()
        if Path(self.absolute_test_file).exists():
            Path(self.absolute_test_file).unlink()
        shutil.rmtree(self.sub_directory, ignore_errors=True)

    def test_path_from_string_absolute(self):
        result = get_path(str(self.absolute_test_file))
        self.assertEqual(result, self.absolute_test_file)

    def test_path_from_string_relative(self):
        result = get_path(self.relative_test_file)
        self.assertEqual(result, Path(self.relative_test_file)
                         )  # Comparing relative paths

    def test_path_from_path_object_absolute(self):
        result = get_path(self.absolute_test_file)
        self.assertEqual(result, self.absolute_test_file)

    def test_path_from_path_object_relative(self):
        result = get_path(Path(self.relative_test_file))
        self.assertEqual(result, Path(self.relative_test_file)
                         )  # Comparing relative paths

    def test_non_existent_path(self):
        with self.assertRaises(FileNotFoundError):
            get_path('nonexistentfile.txt')


if __name__ == "__main__":
    unittest.main()



import unittest
import os

import numpy as np
import openmdao.api as om
from openmdao.utils.assert_utils import (assert_check_partials,
                                         assert_near_equal)

from aviary.mission.gasp_based.flight_conditions import FlightConditions
from aviary.variable_info.enums import SpeedType
from aviary.variable_info.variables import Dynamic


class FlightConditionsTestCase1(unittest.TestCase):
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "group",
            FlightConditions(num_nodes=2, input_speed_type=SpeedType.TAS),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(
            "rho", val=1.22 * np.ones(2), units="kg/m**3"
        )
        self.prob.model.set_input_defaults(
            Dynamic.Mission.SPEED_OF_SOUND,
            val=344 * np.ones(2),
            units="m/s")
        self.prob.model.set_input_defaults("TAS", val=344 * np.ones(2), units="m/s")

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):
        tol = 1e-5
        self.prob.run_model()

        assert_near_equal(
            self.prob[Dynamic.Mission.DYNAMIC_PRESSURE], 1507.6 * np.ones(2), tol)
        assert_near_equal(self.prob[Dynamic.Mission.MACH], np.ones(2), tol)
        assert_near_equal(
            self.prob.get_val("EAS", units="m/s"), 343.3 * np.ones(2), tol
        )

        partial_data = self.prob.check_partials(out_stream=None, method="cs")

        assert_check_partials(partial_data, atol=1e-8, rtol=1e-8)


class FlightConditionsTestCase2(unittest.TestCase):
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "group",
            FlightConditions(num_nodes=2, input_speed_type=SpeedType.EAS),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(
            "rho", val=1.05 * np.ones(2), units="kg/m**3"
        )
        self.prob.model.set_input_defaults(
            Dynamic.Mission.SPEED_OF_SOUND,
            val=344 * np.ones(2),
            units="m/s")
        self.prob.model.set_input_defaults(
            "EAS", val=318.4821143 * np.ones(2), units="m/s"
        )

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):
        tol = 1e-5
        self.prob.run_model()

        assert_near_equal(
            self.prob[Dynamic.Mission.DYNAMIC_PRESSURE], 1297.54 * np.ones(2), tol)
        assert_near_equal(self.prob["TAS"], 1128.61 * np.ones(2), tol)
        assert_near_equal(self.prob[Dynamic.Mission.MACH], np.ones(2), tol)

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-8, rtol=1e-8)


class FlightConditionsTestCase3(unittest.TestCase):
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "group",
            FlightConditions(num_nodes=2, input_speed_type=SpeedType.MACH),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(
            "rho", val=1.05 * np.ones(2), units="kg/m**3"
        )
        self.prob.model.set_input_defaults(
            Dynamic.Mission.SPEED_OF_SOUND,
            val=344 * np.ones(2),
            units="m/s")
        self.prob.model.set_input_defaults(
            Dynamic.Mission.MACH, val=np.ones(2), units="unitless")

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):
        tol = 1e-5
        self.prob.run_model()

        assert_near_equal(
            self.prob[Dynamic.Mission.DYNAMIC_PRESSURE], 1297.54 * np.ones(2), tol)
        assert_near_equal(self.prob["TAS"], 1128.61 * np.ones(2), tol)
        assert_near_equal(
            self.prob.get_val("EAS", units="m/s"), 318.4821143 * np.ones(2), tol
        )

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-8, rtol=1e-8)


if __name__ == "__main__":
    unittest.main()



from aviary.mission.phase_builder_base import PhaseBuilderBase
from aviary.mission.initial_guess_builders import InitialGuessState, InitialGuessTime, InitialGuessControl
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.variables import Dynamic
from aviary.mission.gasp_based.ode.groundroll_ode import GroundrollODE


class GroundrollPhase(PhaseBuilderBase):
    default_name = 'groundroll_phase'
    default_ode_class = GroundrollODE

    _meta_data_ = {}
    _initial_guesses_meta_data_ = {}

    def build_phase(self, aviary_options: AviaryValues = None):
        phase = self.phase = super().build_phase(aviary_options)

        # Retrieve user options values
        user_options = self.user_options
        # Add the necessary get_val calls for each parameter, e.g.,
        fix_initial = user_options.get_val('fix_initial')
        fix_initial_mass = user_options.get_val('fix_initial_mass')
        connect_initial_mass = user_options.get_val('connect_initial_mass')
        mass_lower = user_options.get_val('mass_lower', units='lbm')
        mass_upper = user_options.get_val('mass_upper', units='lbm')
        mass_ref = user_options.get_val('mass_ref', units='lbm')
        mass_ref0 = user_options.get_val('mass_ref0', units='lbm')
        mass_defect_ref = user_options.get_val('mass_defect_ref', units='lbm')
        distance_lower = user_options.get_val('distance_lower', units='ft')
        distance_upper = user_options.get_val('distance_upper', units='ft')
        distance_ref = user_options.get_val('distance_ref', units='ft')
        distance_ref0 = user_options.get_val('distance_ref0', units='ft')
        distance_defect_ref = user_options.get_val('distance_defect_ref', units='ft')

        # Add states
        self.add_velocity_state(user_options)

        phase.add_state(
            Dynamic.Mission.MASS,
            fix_initial=fix_initial_mass,
            input_initial=connect_initial_mass,
            fix_final=False,
            lower=mass_lower,
            upper=mass_upper,
            units="lbm",
            rate_source=Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL,
            ref=mass_ref,
            defect_ref=mass_defect_ref,
            ref0=mass_ref0,
        )

        phase.add_state(
            Dynamic.Mission.DISTANCE,
            fix_initial=fix_initial,
            fix_final=False,
            lower=distance_lower,
            upper=distance_upper,
            units="ft",
            rate_source="distance_rate",
            ref=distance_ref,
            defect_ref=distance_defect_ref,
            ref0=distance_ref0,
        )

        phase.add_parameter("t_init_gear", units="s",
                            static_target=True, opt=False, val=100)
        phase.add_parameter("t_init_flaps", units="s",
                            static_target=True, opt=False, val=100)

        # boundary/path constraints + controls
        # the final TAS is constrained externally to define the transition to the rotation
        # phase

        phase.add_timeseries_output(Dynamic.Mission.THRUST_TOTAL, units="lbf")

        phase.add_timeseries_output("normal_force")
        phase.add_timeseries_output(Dynamic.Mission.MACH)
        phase.add_timeseries_output("EAS", units="kn")

        phase.add_timeseries_output(Dynamic.Mission.LIFT)
        phase.add_timeseries_output("CL")
        phase.add_timeseries_output("CD")
        phase.add_timeseries_output("fuselage_pitch", output_name="theta", units="deg")

        return phase


# Adding metadata for the GroundrollPhase
GroundrollPhase._add_meta_data('fix_initial', val=True)
GroundrollPhase._add_meta_data('fix_initial_mass', val=False)
GroundrollPhase._add_meta_data('connect_initial_mass', val=True)
GroundrollPhase._add_meta_data('duration_bounds', val=(1, 100), units='s')
GroundrollPhase._add_meta_data('duration_ref', val=1, units='s')
GroundrollPhase._add_meta_data('velocity_lower', val=0, units='kn')
GroundrollPhase._add_meta_data('velocity_upper', val=1000, units='kn')
GroundrollPhase._add_meta_data('velocity_ref', val=100, units='kn')
GroundrollPhase._add_meta_data('velocity_ref0', val=0, units='kn')
GroundrollPhase._add_meta_data('velocity_defect_ref', val=None, units='kn')
GroundrollPhase._add_meta_data('mass_lower', val=0, units='lbm')
GroundrollPhase._add_meta_data('mass_upper', val=200_000, units='lbm')
GroundrollPhase._add_meta_data('mass_ref', val=100_000, units='lbm')
GroundrollPhase._add_meta_data('mass_ref0', val=0, units='lbm')
GroundrollPhase._add_meta_data('mass_defect_ref', val=100, units='lbm')
GroundrollPhase._add_meta_data('distance_lower', val=0, units='ft')
GroundrollPhase._add_meta_data('distance_upper', val=4000, units='ft')
GroundrollPhase._add_meta_data('distance_ref', val=3000, units='ft')
GroundrollPhase._add_meta_data('distance_ref0', val=0, units='ft')
GroundrollPhase._add_meta_data('distance_defect_ref', val=3000, units='ft')
GroundrollPhase._add_meta_data('t_init_gear', val=100, units='s')
GroundrollPhase._add_meta_data('t_init_flaps', val=100, units='s')
GroundrollPhase._add_meta_data('num_segments', val=None, units='unitless')
GroundrollPhase._add_meta_data('order', val=None, units='unitless')

# Adding initial guess metadata
GroundrollPhase._add_initial_guess_meta_data(
    InitialGuessTime(),
    desc='initial guess for time options')
GroundrollPhase._add_initial_guess_meta_data(
    InitialGuessState('velocity'),
    desc='initial guess for true airspeed state')
GroundrollPhase._add_initial_guess_meta_data(
    InitialGuessState('mass'),
    desc='initial guess for mass state')
GroundrollPhase._add_initial_guess_meta_data(
    InitialGuessState('distance'),
    desc='initial guess for distance state')
GroundrollPhase._add_initial_guess_meta_data(
    InitialGuessControl('throttle'),
    desc='initial guess for throttle')



import openmdao.api as om

from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft


class TotalWettedArea(om.ExplicitComponent):

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.Canard.WETTED_AREA, 0.0)
        add_aviary_input(self, Aircraft.Fuselage.WETTED_AREA, 0.0)
        add_aviary_input(self, Aircraft.HorizontalTail.WETTED_AREA, 0.0)
        add_aviary_input(self, Aircraft.Nacelle.TOTAL_WETTED_AREA, 0.0)
        add_aviary_input(self, Aircraft.VerticalTail.WETTED_AREA, 0.0)
        add_aviary_input(self, Aircraft.Wing.WETTED_AREA, 0.0)

        add_aviary_output(self, Aircraft.Design.TOTAL_WETTED_AREA, 0.0)

    def setup_partials(self):
        self.declare_partials('*', '*', val=1.0)

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        outputs[Aircraft.Design.TOTAL_WETTED_AREA] = (
            inputs[Aircraft.Canard.WETTED_AREA]
            + inputs[Aircraft.Fuselage.WETTED_AREA]
            + inputs[Aircraft.HorizontalTail.WETTED_AREA]
            + inputs[Aircraft.Nacelle.TOTAL_WETTED_AREA]
            + inputs[Aircraft.VerticalTail.WETTED_AREA]
            + inputs[Aircraft.Wing.WETTED_AREA])



"""
Define subsystem builder for Aviary core geometry.

Classes
-------
GeometryBuilderBase: the interface for a geometry subsystem builder.

CoreGeometryBuilder : the interface for Aviary's core geometry subsystem builder
"""

from aviary.interface.utils.markdown_utils import write_markdown_variable_table
from aviary.subsystems.subsystem_builder_base import SubsystemBuilderBase
from aviary.subsystems.geometry.combined_geometry import CombinedGeometry
from aviary.subsystems.geometry.flops_based.prep_geom import PrepGeom
from aviary.subsystems.geometry.gasp_based.size_group import SizeGroup
from aviary.variable_info.variables import Aircraft
from aviary.variable_info.enums import LegacyCode


GASP = LegacyCode.GASP
FLOPS = LegacyCode.FLOPS

_default_name = 'geometry'


class GeometryBuilderBase(SubsystemBuilderBase):
    def __init__(self, name=None, meta_data=None):
        if name is None:
            name = _default_name

        super().__init__(name=name, meta_data=meta_data)

    def mission_inputs(self, **kwargs):
        return ['*']

    def mission_outputs(self, **kwargs):
        return ['*']


class CoreGeometryBuilder(GeometryBuilderBase):
    def __init__(self, name=None, meta_data=None, code_origin=None,
                 use_both_geometries=False, code_origin_to_prioritize=None):
        if name is None:
            name = 'core_geometry'

        if code_origin not in (FLOPS, GASP) and not use_both_geometries:
            raise ValueError('Code origin is not one of the following: (FLOPS, GASP)')

        self.code_origin = code_origin
        self.use_both_geometries = use_both_geometries
        self.code_origin_to_prioritize = code_origin_to_prioritize

        super().__init__(name=name, meta_data=meta_data)

    def build_pre_mission(self, aviary_inputs):
        code_origin = self.code_origin
        both_geom = self.use_both_geometries
        code_origin_to_prioritize = self.code_origin_to_prioritize

        geom_group = None

        if both_geom:
            geom_group = CombinedGeometry(aviary_options=aviary_inputs,
                                          code_origin_to_prioritize=code_origin_to_prioritize)

        elif code_origin is GASP:
            geom_group = SizeGroup(aviary_options=aviary_inputs)
            geom_group.manual_overrides = None

        elif code_origin is FLOPS:
            geom_group = PrepGeom(aviary_options=aviary_inputs)
            geom_group.manual_overrides = None

        return geom_group

    def build_mission(self, num_nodes, aviary_inputs, **kwargs):
        super().build_mission(num_nodes, aviary_inputs)

    def report(self, prob, reports_folder, **kwargs):
        """
        Generate the report for Aviary core geometry analysis

        Parameters
        ----------
        prob : AviaryProblem
            The AviaryProblem that will be used to generate the report
        reports_folder : Path
            Location of the subsystems_report folder this report will be placed in
        """
        filename = self.name + '.md'
        filepath = reports_folder / filename

        # TODO output differs by method
        # TODO finish variables of interest
        wing_outputs = [Aircraft.Wing.AREA,
                        Aircraft.Wing.SPAN,
                        Aircraft.Wing.ASPECT_RATIO,
                        Aircraft.Wing.SWEEP]
        htail_outputs = [Aircraft.HorizontalTail.AREA,
                         Aircraft.VerticalTail.AREA]
        fuselage_outputs = [Aircraft.Fuselage.LENGTH,
                            Aircraft.Fuselage.AVG_DIAMETER]

        with open(filepath, mode='w') as f:
            method = self.code_origin + ' METHOD'
            if self.use_both_geometries:
                method = ('FLOPS AND GASP METHODS')
            f.write(f'# GEOMETRY: {method}\n')
            f.write('## Wing')
            write_markdown_variable_table(f, prob, wing_outputs, self.meta_data)
            f.write('\n## Empennage\n')
            f.write('### Horizontal Tail')
            write_markdown_variable_table(f, prob, htail_outputs, self.meta_data)
            f.write('\n## Fuselage')
            write_markdown_variable_table(f, prob, fuselage_outputs, self.meta_data)



import numpy as np
import openmdao.api as om

import aviary.constants as constants
from aviary.variable_info.functions import add_aviary_input
from aviary.variable_info.variables import Aircraft, Dynamic

grav_metric = constants.GRAV_METRIC_FLOPS


class SimpleLift(om.ExplicitComponent):
    '''
    Calculate lift as a function of wing area, dynamic pressure, and lift coefficient.
    '''

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        add_aviary_input(self, Aircraft.Wing.AREA, val=1., units='m**2')

        self.add_input(
            Dynamic.Mission.DYNAMIC_PRESSURE, val=np.ones(nn), units='N/m**2',
            desc='pressure caused by fluid motion')

        self.add_input(
            name='cl', val=np.ones(nn), desc='current coefficient of lift',
            units='unitless')

        self.add_output(name=Dynamic.Mission.LIFT,
                        val=np.ones(nn), desc='Lift', units='N')

    def setup_partials(self):
        nn = self.options['num_nodes']
        rows_cols = np.arange(nn)

        self.declare_partials(Dynamic.Mission.LIFT, Aircraft.Wing.AREA)

        self.declare_partials(
            Dynamic.Mission.LIFT, [Dynamic.Mission.DYNAMIC_PRESSURE, 'cl'], rows=rows_cols, cols=rows_cols)

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        S = inputs[Aircraft.Wing.AREA]
        q = inputs[Dynamic.Mission.DYNAMIC_PRESSURE]
        CL = inputs['cl']

        outputs[Dynamic.Mission.LIFT] = q * S * CL

    def compute_partials(self, inputs, partials, discrete_inputs=None):
        S = inputs[Aircraft.Wing.AREA]
        q = inputs[Dynamic.Mission.DYNAMIC_PRESSURE]
        CL = inputs['cl']

        partials[Dynamic.Mission.LIFT, Aircraft.Wing.AREA] = q * CL
        partials[Dynamic.Mission.LIFT, Dynamic.Mission.DYNAMIC_PRESSURE] = S * CL
        partials[Dynamic.Mission.LIFT, 'cl'] = q * S


class LiftEqualsWeight(om.ExplicitComponent):
    """
    Compute lift given aircraft mass.

    This is valid for non-accelerating phases only.
    """

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        add_aviary_input(self, varname=Aircraft.Wing.AREA, val=1.0, units='m**2')

        self.add_input(
            name=Dynamic.Mission.MASS, val=np.ones(nn), desc='current aircraft mass',
            units='kg')

        self.add_input(
            Dynamic.Mission.DYNAMIC_PRESSURE, val=np.ones(nn), units='N/m**2',
            desc='pressure caused by fluid motion')

        self.add_output(
            name='cl', val=np.ones(nn), desc='current coefficient of lift',
            units='unitless')

        self.add_output(name=Dynamic.Mission.LIFT,
                        val=np.ones(nn), desc='Lift', units='N')

    def setup_partials(self):
        nn = self.options['num_nodes']
        row_col = np.arange(nn)

        self.declare_partials(
            Dynamic.Mission.LIFT, Dynamic.Mission.MASS, rows=row_col, cols=row_col, val=grav_metric)

        self.declare_partials(
            Dynamic.Mission.LIFT, [Aircraft.Wing.AREA, Dynamic.Mission.DYNAMIC_PRESSURE], dependent=False)

        self.declare_partials('cl', Aircraft.Wing.AREA)

        self.declare_partials(
            'cl', [Dynamic.Mission.MASS, Dynamic.Mission.DYNAMIC_PRESSURE], rows=row_col, cols=row_col)

    def compute(self, inputs, outputs):
        S = inputs[Aircraft.Wing.AREA]
        q = inputs[Dynamic.Mission.DYNAMIC_PRESSURE]
        weight = grav_metric * inputs[Dynamic.Mission.MASS]

        outputs['cl'] = weight / (q * S)

        outputs[Dynamic.Mission.LIFT] = weight

    def compute_partials(self, inputs, partials, discrete_inputs=None):
        S = inputs[Aircraft.Wing.AREA]
        q = inputs[Dynamic.Mission.DYNAMIC_PRESSURE]
        weight = grav_metric * inputs[Dynamic.Mission.MASS]

        f = weight / q
        # df = 0.
        g = S
        # dg = 1.
        partials['cl', Aircraft.Wing.AREA] = -f / g**2

        partials['cl', Dynamic.Mission.MASS] = grav_metric / (q * S)

        f = weight / S
        # df = 0.
        g = q
        # dg = 1.
        partials['cl', Dynamic.Mission.DYNAMIC_PRESSURE] = -f / g**2



import warnings

import openmdao.api as om

from aviary.interface.default_phase_info.two_dof_fiti import create_2dof_based_descent_phases
from aviary.mission.gasp_based.phases.time_integration_traj import FlexibleTraj
from aviary.variable_info.variables import Aircraft, Mission, Dynamic


def descent_range_and_fuel(
    phases=None,
    ode_args=None,
    initial_mass=154e3,
    cruise_alt=35e3,
    cruise_mach=.8,
    empty_weight=85e3,
    payload_weight=30800,
    reserve_fuel=4998,
):

    prob = om.Problem()
    prob.driver = om.pyOptSparseDriver()
    prob.driver.options["optimizer"] = 'IPOPT'
    prob.driver.opt_settings['tol'] = 1.0E-6
    prob.driver.opt_settings['mu_init'] = 1e-5
    prob.driver.opt_settings['max_iter'] = 50
    prob.driver.opt_settings['print_level'] = 5

    if phases is None:
        phases = create_2dof_based_descent_phases(
            ode_args,
            cruise_mach=cruise_mach,
        )

    traj = FlexibleTraj(
        Phases=phases,
        traj_final_state_output=[
            Dynamic.Mission.MASS,
            Dynamic.Mission.DISTANCE,
            Dynamic.Mission.ALTITUDE,
        ],
        traj_initial_state_input=[
            Dynamic.Mission.MASS,
            Dynamic.Mission.DISTANCE,
            Dynamic.Mission.ALTITUDE,
        ],
    )
    prob.model = om.Group()
    prob.model.add_subsystem('traj', traj)

    prob.model.add_subsystem(
        "fuel_obj",
        om.ExecComp(
            "reg_objective = overall_fuel/10000",
            reg_objective={"val": 0.0, "units": "unitless"},
            overall_fuel={"units": "lbm"},
        ),
        promotes_inputs=[
            ("overall_fuel", Mission.Summary.TOTAL_FUEL_MASS),
        ],
        promotes_outputs=[("reg_objective", Mission.Objectives.FUEL)],
    )

    prob.model.add_objective(Mission.Objectives.FUEL, ref=1e4)

    prob.setup()
    prob.set_val("traj.altitude_initial", val=cruise_alt, units="ft")
    prob.set_val("traj.mass_initial", val=initial_mass, units="lbm")
    prob.set_val("traj.distance_initial", val=0, units="NM")

    # prevent UserWarning that is displayed when an event is triggered
    warnings.filterwarnings('ignore', category=UserWarning)
    prob.run_model()
    warnings.filterwarnings('default', category=UserWarning)

    final_range = prob.get_val('traj.distance_final', units='NM')[0]
    final_mass = prob.get_val('traj.mass_final', units='lbm')[0]
    descent_fuel = initial_mass - final_mass
    print('final range: ', final_range)
    print('fuel burned: ', descent_fuel)

    initial_mass2 = empty_weight + payload_weight + reserve_fuel + descent_fuel
    prob.set_val("traj.mass_initial", val=initial_mass2, units="lbm")

    # prevent UserWarning that is displayed when an event is triggered
    warnings.filterwarnings('ignore', category=UserWarning)
    prob.run_model()
    warnings.filterwarnings('default', category=UserWarning)

    final_range2 = prob.get_val('traj.distance_final', units='NM')[0]
    final_mass2 = prob.get_val('traj.mass_final', units='lbm')[0]
    descent_fuel2 = initial_mass2 - final_mass2
    print('initial mass: ', initial_mass2)
    print('final range: ', final_range2)
    print('fuel burned: ', initial_mass2-final_mass2)

    results = {
        'initial_guess': {
            'distance_flown': final_range,
            'fuel_burned': descent_fuel
        },
        'refined_guess': {
            'distance_flown': final_range2,
            'fuel_burned': descent_fuel2
        }
    }

    return results



import numpy as np
import openmdao.api as om
from dymos.models.atmosphere import USatm1976Comp

from aviary.mission.flops_based.ode.mission_EOM import MissionEOM
from aviary.utils.aviary_values import AviaryValues
from aviary.utils.functions import promote_aircraft_and_mission_vars
from aviary.variable_info.variable_meta_data import _MetaData
from aviary.variable_info.variables import Dynamic, Mission
from aviary.variable_info.variables_in import VariablesIn


class ExternalSubsystemGroup(om.Group):
    def configure(self):
        promote_aircraft_and_mission_vars(self)


class MissionODE(om.Group):
    def initialize(self):
        self.options.declare(
            'num_nodes', types=int,
            desc='Number of nodes to be evaluated in the RHS')
        self.options.declare(
            'subsystem_options', types=dict, default={},
            desc='dictionary of parameters to be passed to the subsystem builders')
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')
        self.options.declare(
            'core_subsystems',
            desc='list of core subsystem builder instances to be added to the ODE'
        )
        self.options.declare(
            'external_subsystems', default=[],
            desc='list of external subsystem builder instances to be added to the ODE')
        self.options.declare(
            'meta_data', default=_MetaData,
            desc='metadata associated with the variables to be passed into the ODE')
        self.options.declare(
            'use_actual_takeoff_mass', default=False,
            desc='flag to use actual takeoff mass in the climb phase, otherwise assume 100 kg fuel burn')
        self.options.declare(
            'throttle_enforcement', default='path_constraint',
            values=['path_constraint', 'boundary_constraint', 'bounded', None],
            desc='flag to enforce throttle constraints on the path or at the segment boundaries or using solver bounds'
        )

    def setup(self):
        options = self.options
        nn = options['num_nodes']
        aviary_options = options['aviary_options']
        core_subsystems = options['core_subsystems']
        subsystem_options = options['subsystem_options']
        engine_count = len(aviary_options.get_val('engine_models'))

        self.add_subsystem(
            'input_port',
            VariablesIn(aviary_options=aviary_options,
                        meta_data=self.options['meta_data'],
                        context='mission'),
            promotes_inputs=['*'],
            promotes_outputs=['*'])
        self.add_subsystem(
            name='atmosphere',
            subsys=USatm1976Comp(num_nodes=nn),
            promotes_inputs=[('h', Dynamic.Mission.ALTITUDE)],
            promotes_outputs=[
                ('sos', Dynamic.Mission.SPEED_OF_SOUND), ('rho', Dynamic.Mission.DENSITY),
                ('temp', Dynamic.Mission.TEMPERATURE), ('pres', Dynamic.Mission.STATIC_PRESSURE)])

        # add an execcomp to compute velocity based off mach and sos
        self.add_subsystem(
            name='velocity_comp',
            subsys=om.ExecComp(
                'velocity = mach * sos',
                mach={'units': 'unitless', 'shape': (nn,)},
                sos={'units': 'm/s', 'shape': (nn,)},
                velocity={'units': 'm/s', 'shape': (nn,)},
                has_diag_partials=True,
            ),
            promotes_inputs=[('mach', Dynamic.Mission.MACH),
                             ('sos', Dynamic.Mission.SPEED_OF_SOUND)],
            promotes_outputs=[('velocity', Dynamic.Mission.VELOCITY)],
        )

        # add execcomp to compute velocity_rate based off mach_rate and sos
        self.add_subsystem(
            name='velocity_rate_comp',
            subsys=om.ExecComp(
                'velocity_rate = mach_rate * sos',
                mach_rate={'units': 'unitless/s', 'shape': (nn,)},
                sos={'units': 'm/s', 'shape': (nn,)},
                velocity_rate={'units': 'm/s**2', 'shape': (nn,)},
                has_diag_partials=True,
            ),
            promotes_inputs=[('mach_rate', Dynamic.Mission.MACH_RATE),
                             ('sos', Dynamic.Mission.SPEED_OF_SOUND)],
            promotes_outputs=[('velocity_rate', Dynamic.Mission.VELOCITY_RATE)],
        )

        base_options = {'num_nodes': nn, 'aviary_inputs': aviary_options}

        for subsystem in core_subsystems:
            # check if subsystem_options has entry for a subsystem of this name
            if subsystem.name in subsystem_options:
                kwargs = subsystem_options[subsystem.name]
            else:
                kwargs = {}

            kwargs.update(base_options)
            system = subsystem.build_mission(**kwargs)

            if system is not None:
                self.add_subsystem(subsystem.name,
                                   system,
                                   promotes_inputs=subsystem.mission_inputs(**kwargs),
                                   promotes_outputs=subsystem.mission_outputs(**kwargs))

        # Create a lightly modified version of an OM group to add external subsystems
        # to the ODE with a special configure() method that promotes
        # all aircraft:* and mission:* variables to the ODE.
        external_subsystem_group = ExternalSubsystemGroup()
        add_subsystem_group = False

        for subsystem in self.options['external_subsystems']:
            subsystem_mission = subsystem.build_mission(
                num_nodes=nn, aviary_inputs=aviary_options)
            if subsystem_mission is not None:
                add_subsystem_group = True
                external_subsystem_group.add_subsystem(subsystem.name, subsystem_mission)

        # Only add the external subsystem group if it has at least one subsystem.
        # Without this logic there'd be an empty OM group added to the ODE.
        if add_subsystem_group:
            self.add_subsystem(
                name='external_subsystems',
                subsys=external_subsystem_group,
                promotes_inputs=['*'],
                promotes_outputs=['*'])

        self.add_subsystem(
            name='mission_EOM',
            subsys=MissionEOM(num_nodes=nn),
            promotes_inputs=[
                Dynamic.Mission.VELOCITY, Dynamic.Mission.MASS,
                Dynamic.Mission.THRUST_MAX_TOTAL,
                Dynamic.Mission.DRAG, Dynamic.Mission.ALTITUDE_RATE,
                Dynamic.Mission.VELOCITY_RATE],
            promotes_outputs=[
                Dynamic.Mission.SPECIFIC_ENERGY_RATE_EXCESS,
                Dynamic.Mission.ALTITUDE_RATE_MAX,
                Dynamic.Mission.DISTANCE_RATE,
                'thrust_required',
            ])

        # add a balance comp to compute throttle based on the altitude rate
        self.add_subsystem(name='throttle_balance',
                           subsys=om.BalanceComp(name=Dynamic.Mission.THROTTLE,
                                                 units="unitless",
                                                 val=np.ones(nn),
                                                 lhs_name='thrust_required',
                                                 rhs_name=Dynamic.Mission.THRUST_TOTAL,
                                                 eq_units="lbf",
                                                 normalize=False,
                                                 lower=0.0 if options['throttle_enforcement'] == 'bounded' else None,
                                                 upper=1.0 if options['throttle_enforcement'] == 'bounded' else None,
                                                 res_ref=1.0e6,
                                                 ),
                           promotes_inputs=['*'],
                           promotes_outputs=['*'])

        self.set_input_defaults(Dynamic.Mission.MASS, val=np.ones(nn), units='kg')
        self.set_input_defaults(Dynamic.Mission.VELOCITY, val=np.ones(nn), units='m/s')
        self.set_input_defaults(Dynamic.Mission.ALTITUDE, val=np.ones(nn), units='m')
        self.set_input_defaults(
            Dynamic.Mission.THROTTLE, val=np.ones((nn, engine_count)),
            units='unitless')

        if options['use_actual_takeoff_mass']:
            exec_comp_string = 'initial_mass_residual = initial_mass - mass[0]'
            initial_mass_string = Mission.Takeoff.FINAL_MASS
        else:
            exec_comp_string = 'initial_mass_residual = initial_mass - mass[0] - 100.'
            initial_mass_string = Mission.Design.GROSS_MASS

        # Experimental: Add a component to constrain the initial mass to be equal to design gross weight.
        initial_mass_residual_constraint = om.ExecComp(
            exec_comp_string,
            initial_mass={'units': 'kg'},
            mass={'units': 'kg', 'shape': (nn,)},
            initial_mass_residual={'units': 'kg', 'res_ref': 1.0e5},
        )

        self.add_subsystem('initial_mass_residual_constraint', initial_mass_residual_constraint,
                           promotes_inputs=[
                               ('initial_mass', initial_mass_string),
                               ('mass', Dynamic.Mission.MASS)
                           ],
                           promotes_outputs=['initial_mass_residual'])

        self.nonlinear_solver = om.NewtonSolver(solve_subsystems=True,
                                                atol=1.0e-10,
                                                rtol=1.0e-10,
                                                )
        self.nonlinear_solver.linesearch = om.BoundsEnforceLS()
        self.linear_solver = om.DirectSolver(assemble_jac=True)
        self.nonlinear_solver.options['err_on_non_converge'] = True
        self.nonlinear_solver.options['iprint'] = 2
